#!/usr/bin/env python
import numpy as np
import pyjams.const as const


__all__ = ['cuntz_gleixner']


def cuntz_gleixner(idecdate, iGPP, iRd, iCa, iRa, igtot, sunrise, Vcyt=None,
                   date0=False,
                   V0starch=const.eps,
                   R0starch=const.R13VPDB,
                   R0cyt=const.R13VPDB,
                   daynight=None, daylength=57600,
                   Phi=0.3, s_resid=const.eps,
                   betas=None, betap=0.75,
                   epsa=4.4e-3, epsb=29.5e-3,
                   epsg=20.0e-3, epst=-4.4e-3,
                   epss=10.0e-3, epsp=1.0e-3,
                   steady=False,
                   Rass=False, Rm=False, Rchl=False,
                   Rcyt=False, Rstarch=False,
                   Rpyr=False, Rbio=False,
                   Rphloem=False,
                   Vstarch=False, ass13=False, disc=False,
                   Rnew_starch=False, Rnew_cyt=False,
                   fullmodel=True, julian=True, nocheck=False,
                   starch_mol2g=None, V0starchg=const.eps):
    """
    Calculates the Cuntz-Gleixner steady state and non-steady state models
    of 13C discrimiantion in the Calvin cycle.


    Definition
    ----------
    def cuntz_gleixner(idecdate, iGPP, iRd, iCa, iRa, igtot, sunrise,
                       Vcyt=None,
                       date0=False,
                       V0starch=const.eps, R0starch=const.R13VPDB,
                       R0cyt=const.R13VPDB,
                       daynight=None, daylength=57600,
                       Phi=0.3, s_resid=const.eps,
                       betas=None, betap=0.75,
                       epsa=4.4e-3, epsb=29.5e-3,
                       epsg=20.0e-3, epst=-4.4e-3,
                       epss=10.0e-3, epsp=1.0e-3,
                       steady=False,
                       Rass=False, Rm=False, Rchl=False,
                       Rcyt=False, Rstarch=False,
                       Rpyr=False, Rbio=False,
                       Rphloem=False,
                       Vstarch=False, ass13=False, disc=False,
                       Rnew_starch=False, Rnew_cyt=False,
                       fullmodel=True, julian=True, nocheck=False,
                       starch_mol2g=None, V0starchg=const.eps):


    Input
    -----
    idecdate
        decimal date
    iGPP
        Gross Photosynthesis: GPP = A + Rd [umol(CO2)/m2s]
    iRd
        Leaf respiration: GPP - A [umol(CO2)/m2s]
    iCa
        Outside CO2 concentration [ppm=umol(CO2)/umol(air)]
    iRa
        13C/12C ratio of outside CO2 concentration
    igtot
        Total conductance for CO2 from outside air to chloroplast
        [mol(CO2)/m2s]
    sunrise
        decimal date of first sunrise in data set


    Input (only nss model)
    -----
    Vcyt
        C-concentration of sucrose pool in cytoplasm [umol(C)/m2(leaf)]


    Optional Input
    --------------
    date0
        Start date of 1st time step (default: False)
        If False, take same time step as first time step in idecdate
    V0starch
        Initial C-concentration in Starch [umol(C)/m2(leaf)] (default: 1e-6)
    V0starchg
        Initial C-concentration in Starch [g(C)/gDW] (default: 1e-6)
    starch_mol2g
        Conversion factor from [umol(C)/m2(leaf)] to [g(C)/gDW] used for
        starch (default: None)
        If given and Vstarch==True then Vstarchg will be returned as well
    R0starch
        Initial 13C/12C ratio of starch pool (default: PDB)
    R0cyt
        Initial 13C/12C ratio of C in cytoplasm (default: PDB)
    daynight
        1/0 array of day or night (default: False)
        If False, day is when gpp>0
    daylength
        length of daylight [s] (default: 57600 = 16h)
    Phi
        Vc/Vo, ratio of carboxylation to oxygenation of Rudisco (default: 0.3)
    s_resid
        Residual starch concentration at end of night [umol(C)/m2(leaf)]
        (default: 1e-6)
    betas
        factor of leaf respiration transferred to biosynthesis (default: False)
        If False, betas is 3*gpp/max(gpp)
        Note: betas*(1-betap) <= 1: if betap=2/3
              -> betas<3: if betap=5/6 -> betas < 6
    betap
        fraction of respiration occurring during biosynthesis:
        min=2/3; max=4/5 (default: 3/4)
    epsa
        effective fractionation along gtot (default: 4.4e-3)
    epsb
        fractionation of Rubisco (default: 29.5e-3)
    epsg
        fractionation of photorespiration (default: 20e-3)
    epst
        equilibrium fractionation value for starch synthesis (default: -4.4e-3)
    epss
        fractionation of biosynthesis production (default: 10e-3)
    epsp
        fractionation of biosynthesis bifurcation (default: 1e-3)


    Parameter
    ---------
    steady
        If True, steady-state instead of non-steady-state model
        (default: False)
    Rass
        If True, output 13C/12C ratio of assimilated carbon (default: False)
    Rm
        If True, output 13C/12C ratio of chloroplast CO2 (default: False)
    Rchl
        If True, output 13C/12C ratio of sugars in chloroplast (default: False)
    Rcyt
        If True, output 13C/12C ratio of sugars in cytoplasm (default: False)
    Rstarch
        If True, output 13C/12C ratio of starch (default: False)
    Rpyr
        If True, output 13C/12C ratio of sugars at pyruvate pathway
        (default: False)
    Rbio
        If True, output 13C/12C ratio of biosynthesis products before
        bifurcation (default: False)
    Rphloem
        If True, output 13C/12C ratio of new phloem products
        (sugars & biosynthesis products) (default: False)
    Vstarch
        If True, output C-concentration in Starch [umol(C)/m2(leaf)]
        (default: False)
    ass13
        If True, output 13C assimilation rate  [umol(13C)/m2s] (default: False)
    disc
        If True, output Discrimination 1-Rass/Ra (default: False)
    Rnew_starch
        If True, output 13C/12C ratio of newly produced starch (default: False)
    Rnew_cyt
        If True, output 13C/12C ratio of newly produced sugars in cytoplasm
        (default: False)
    fullmodel
        If True, output all in the above order (default: True)
    julian
        If True, dates are given as Julian days, otherwise as decimal year
        (default: True)
    nocheck
        If True, do not check betap and betas ranges (default: False)


    Output
    ------
    if fullmodel=True
      Rass, Rm, Rchl, Rcyt, Rstarch, Rpyr, Rbio, Rphloem, Vstarch, ass13, disc,
      Rnew_starch, Rnew_cyt
    and if Vstarch=True and starch_mol2g!=None
      Vstarchg


    Restrictions
    ------------
    If at least one individual output parameter is True then fullmode=False.


    References
    ----------
    Tcherkez G, Farquhar GD, Badeck F & Ghashghaie J,
       Theoretical considerations about carbon isotope distribution in glucose
       of C3 plants, Functional Plant Biology 31, 857-877, 2004
    Gessler A, Tcherkez G, Peuke AD, Ghashghaie J & Farquhar GD,
       Experimental evidence for diel variations of the carbon isotope
       composition in leaf, stem and phloem sap organic matter in Ricinus
       communis, Plant, Cell and Environment 31, 941-953, 2004


    Examples
    --------
    # steady state
    >>> adecdate = np.array([2008.918658925319050,2008.918772768671033,2008.918886612022106,
    ...                      2008.919000455374089,2008.919114298724935,2008.919228142076918,
    ...                      2008.919341985427991,2008.919455828779974,2008.919569672131956,
    ...                      2008.919683515483030,2008.919797358835012,2008.919911202186086,
    ...                      2008.920025045538068,2008.920138888888914,2008.920252732240897,
    ...                      2008.920366575591970,2008.920480418943953,2008.920556314511941,
    ...                      2008.920594262295026,2008.920708105647009,2008.920821948998992,
    ...                      2008.920935792350065,2008.921049635702047,2008.921163479052893,
    ...                      2008.921277322405103,2008.921391165755949,2008.921505009107932])
    >>> gpp = np.array([0.000000000000,23.700827991217,22.449718259243,21.253578109071,20.222525197027,
    ...                   19.503625355216,18.797132965271,18.102416224453,17.780887860470,17.491607940331,
    ...                   17.207072197663,17.089915139494,17.995854647885,18.901914959729,19.681631460738,
    ...                   19.681631460738,19.681631460738,0.000000000000,0.000000000000,0.000000000000,
    ...                   0.000000000000,0.000000000000,0.000000000000,0.000000000000,0.000000000000,
    ...                   0.000000000000] )
    >>> Rd = np.array([0.511900000000,2.361686144687,2.743373026721,3.180029474251,3.476842651940,
    ...                  3.259038512076,3.053641828083,2.860020793216,2.958750580931,3.083603451827,
    ...                  3.213200496886,3.331826587704,3.352936200975,3.374166608865,3.392531460738,
    ...                  3.392531460738,3.392531460738,1.025929405070,0.829676977143,0.633424549217,
    ...                  0.437172122858,0.303515021488,0.547877741613,0.792240464668,1.036603184794,
    ...                  1.280965907360] )
    >>> CO2air = np.array([620.902600000000,537.510500000000,608.806500000000,671.251000000000,
    ...                      652.204000000000,560.157800000000,427.130100000000,395.276000000000,
    ...                      427.000400000000,410.953300000000,386.943500000000,500.417500000000,
    ...                      552.776800000000,515.865800000000,542.450400000000,692.503500000000,
    ...                      656.423500000000,588.844100000000,675.156500000000,725.101900000000,
    ...                      664.837000000000,598.080600000000,610.713600000000,487.087000000000,
    ...                      531.921300000000,675.177700000000] )
    >>> Ra = np.array([0.011067265443,0.011083081802,0.011071245659,0.011060401761,0.011063313320,
    ...                  0.011080216316,0.011111970396,0.011122420992,0.011111174802,0.011116914764,
    ...                  0.011125605614,0.011097923896,0.011079382516,0.011087211473,0.011083896499,
    ...                  0.011057329511,0.011062335683,0.011072518834,0.011061590657,0.011053508863,
    ...                  0.011061281634,0.011071628848,0.011069690431,0.011093962783,0.011086022577,
    ...                  0.011059558971] )
    >>> gtot = np.array([0.064395001124,0.074054920058,0.078085762302,0.078484156864,0.078127160737,
    ...                    0.085209848990,0.088685679784,0.089611189047,0.088528095110,0.086087621579,
    ...                    0.081901616151,0.076984314568,0.080693530135,0.084173028182,0.087005780100,
    ...                    0.087005780100,0.087005780100,0.046798889383,0.042324852911,0.037583815518,
    ...                    0.032460459750,0.028193059760,0.031985237181,0.035564641600,0.038983725824,
    ...                    0.042282334176] )
    >>> ndecdate = 2008.918772768670806
    >>> V0starch = 60498.901260546168
    >>> R0starch = 0.010949362493
    >>> daynight = np.array([0,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,0,0,0,0,0,0,0,0])
    >>> daylength = 56400.
    >>> Phi = np.array([0.081107124487,0.228537911761,0.178664767959,0.161131279216,0.173373369802,
    ...                   0.195820137312,0.276226485398,0.291675965181,0.259480203313,0.285111049806,
    ...                   0.331368464761,0.228636707479,0.199019858196,0.224041328002,0.209848472684,
    ...                   0.147530335706,0.158874834596,0.090595209673,0.077545142771,0.070693599170,
    ...                   0.075277039328,0.082213420279,0.081562178107,0.102842605886,0.095940191411,
    ...                   0.077688517811] )
    >>> s_resid = 40532.561476983901
    >>> betas = np.array([0.000000000000,4.999999000000,4.736061020188,4.483719593421,4.266205481166,
    ...                     4.114544146364,3.965500325995,3.818940732998,3.751110114569,3.690082727626,
    ...                     3.630056187040,3.605340371581,3.796460173119,3.987605459727,4.152096865111,
    ...                     4.152096865111,4.152096865111,0.000000000000,0.000000000000,0.000000000000,
    ...                     0.000000000000,0.000000000000,0.000000000000,0.000000000000,0.000000000000,
    ...                     0.000000000000] )
    >>> betap = 0.8
    >>> epsa = np.array([0.002995512907,0.003039740417,0.003192366495,0.003375544906,0.003479516543,
    ...                    0.003318658489,0.003187204366,0.003078822327,0.003144673535,0.003235857882,
    ...                    0.003343598868,0.003447509132,0.003408830551,0.003373557382,0.003345605523,
    ...                    0.003345605523,0.003345605523,0.003500242606,0.003551923813,0.003615033761,
    ...                    0.003693231875,0.003766948014,0.003706681514,0.003655727827,0.003612282052,
    ...                    0.003574980671] )
    >>> epsb = 0.029
    >>> epsg = 0.0185
    >>> epst = -0.004
    >>> epss = 0.01
    >>> epsp = 0.003
    >>> [Vstarch] = cuntz_gleixner(adecdate[1:], gpp, Rd, CO2air, Ra, gtot, ndecdate,
    ...                            date0=adecdate[0], V0starch=V0starch, R0starch=R0starch, daynight=daynight,
    ...                            daylength=daylength, Phi=Phi, s_resid=s_resid,
    ...                            betas=betas, betap=betap, epsa=epsa, epsb=epsb,
    ...                            epsg=epsg, epst=epst, epss=epss, epsp=epsp,
    ...                            steady=True, Vstarch=True, julian=False)
    >>> from autostring import astr
    >>> print(astr(Vstarch[0:6],5,pp=True))
    [' 40532.56148' ' 70158.59647' ' 98220.74429' '124787.71693' '150065.87342' '174445.40512']
    >>> [Rass,Rm,Rchl,Rcyt,Rstarch,Rpyr,Rbio,Rphloem,Vstarch,ass13,disc,Rnew_starch,Rnew_cyt,Vstarchg] = cuntz_gleixner(
    ...                adecdate[1:], gpp, Rd, CO2air, Ra, gtot, ndecdate,
    ...                date0=adecdate[0], V0starch=V0starch, R0starch=R0starch, daynight=daynight,
    ...                daylength=daylength, Phi=Phi, s_resid=s_resid,
    ...                betas=betas, betap=betap, epsa=epsa, epsb=epsb,
    ...                epsg=epsg, epst=epst, epss=epss, epsp=epsp,
    ...                steady=True, fullmodel=True, julian=False,
    ...                starch_mol2g=1., V0starchg=V0starch)
    >>> print(astr(Rass[0:6],5,pp=True))
    ['0.01095' '0.01092' '0.01087' '0.01083' '0.01083' '0.01085']
    >>> print(astr(Rm[0:6],5,pp=True))
    ['0.01107' '0.01123' '0.01119' '0.01116' '0.01116' '0.01118']
    >>> print(astr(Rchl[0:6],5,pp=True))
    ['0.01095' '0.01092' '0.01087' '0.01084' '0.01084' '0.01086']
    >>> print(astr(Rcyt[0:6],5,pp=True))
    ['0.01095' '0.01092' '0.01087' '0.01084' '0.01084' '0.01086']
    >>> print(astr(Rstarch[0:6],5,pp=True))
    ['0.01095' '0.01095' '0.01094' '0.01093' '0.01092' '0.01092']
    >>> print(astr(Rpyr[0:6],5,pp=True))
    ['0.01095' '0.01105' '0.01100' '0.01097' '0.01097' '0.01099']
    >>> print(astr(Rbio[0:6],5,pp=True))
    ['0.01095' '0.01094' '0.01089' '0.01086' '0.01086' '0.01088']
    >>> print(astr(Rphloem[0:6],5,pp=True))
    ['0.01095' '0.01091' '0.01086' '0.01082' '0.01082' '0.01084']
    >>> print(astr(Vstarch[0:6],5,pp=True))
    [' 40532.56148' ' 70158.59647' ' 98220.74429' '124787.71693' '150065.87342' '174445.40512']
    >>> print(astr(ass13[0:6],5,pp=True))
    ['-0.00560' ' 0.23301' ' 0.21418' ' 0.19580' ' 0.18136' ' 0.17631']
    >>> print(astr(Rnew_starch[0:6],5,pp=True))
    ['0.01095' '0.01096' '0.01091' '0.01088' '0.01088' '0.01090']
    >>> print(astr(Rnew_cyt[0:6],5,pp=True))
    ['0.01095' '0.01092' '0.01087' '0.01084' '0.01084' '0.01086']
    >>> print(astr(Vstarchg[0:6],5,pp=True))
    [' 40532.56148' ' 70158.59647' ' 98220.74429' '124787.71693' '150065.87342' '174445.40512']

    # non-steady state
    >>> R0cyt = 0.010911449304
    >>> Vcyt = np.array([135000.,135000.,135000.,135000.,135000.,135000.,135000.,
    ...                   135000.,135000.,135000.,135000.,135000.,135000.,135000.,
    ...                   135000.,135000.,135000.,135000.,135000.,135000.,135000.,
    ...                   135000.,135000.,135000.,135000.,135000.])
    >>> [Rass,Rm,Rchl,Rcyt,Rstarch,Rpyr,Rbio,Rphloem,Vstarch,ass13,disc,Rnew_starch,Rnew_cyt] = cuntz_gleixner(
    ...                adecdate[1:], gpp, Rd, CO2air, Ra, gtot, ndecdate, Vcyt=Vcyt,
    ...                date0=adecdate[0], V0starch=V0starch, R0starch=R0starch, R0cyt=R0cyt,
    ...                daynight=daynight,
    ...                daylength=daylength, Phi=Phi, s_resid=s_resid,
    ...                betas=betas, betap=betap, epsa=epsa, epsb=epsb,
    ...                epsg=epsg, epst=epst, epss=epss, epsp=epsp,
    ...                steady=False, fullmodel=True, julian=False)
    >>> print(astr(Vstarch[0:6],5,pp=True))
    [' 40532.56148' ' 70158.59647' ' 98220.74429' '124787.71693' '150065.87342' '174445.40512']
    >>> print(astr(ass13[0:6],5,pp=True))
    ['-0.00559' ' 0.23302' ' 0.21413' ' 0.19571' ' 0.18129' ' 0.17629']
    >>> print(astr(Rass[0:6],5,pp=True))
    ['0.01092' '0.01092' '0.01087' '0.01083' '0.01083' '0.01085']
    >>> print(astr(Rm[0:6],5,pp=True))
    ['0.01107' '0.01123' '0.01119' '0.01116' '0.01116' '0.01118']
    >>> print(astr(Rchl[0:6],5,pp=True))
    ['0.01095' '0.01092' '0.01087' '0.01084' '0.01084' '0.01086']
    >>> print(astr(Rstarch[0:6],5,pp=True))
    ['0.01095' '0.01095' '0.01094' '0.01093' '0.01092' '0.01092']
    >>> print(astr(Rcyt[0:6],5,pp=True))
    ['0.01092' '0.01092' '0.01090' '0.01088' '0.01087' '0.01087']
    >>> print(astr(Rpyr[0:6],5,pp=True))
    ['0.01092' '0.01105' '0.01103' '0.01101' '0.01100' '0.01099']
    >>> print(astr(Rbio[0:6],5,pp=True))
    ['0.01092' '0.01094' '0.01092' '0.01090' '0.01089' '0.01088']
    >>> print(astr(Rnew_cyt[0:6],5,pp=True))
    ['0.01095' '0.01092' '0.01087' '0.01084' '0.01084' '0.01086']
    >>> print(astr(Rnew_starch[0:6],5,pp=True))
    ['0.01095' '0.01096' '0.01091' '0.01088' '0.01088' '0.01090']
    >>> print(astr(Rphloem[0:6],5,pp=True))
    ['0.01092' '0.01091' '0.01089' '0.01087' '0.01085' '0.01085']
    >>> from dec2date import dec2date
    >>> from date2dec import date2dec
    >>> aa = dec2date(adecdate, ascii=True, calendar='decimal')
    >>> jadecdate = date2dec(ascii=aa)
    >>> ndecdate = 2008.918772768670806
    >>> bb = dec2date(ndecdate, ascii=True, calendar='decimal')
    >>> jndecdate = date2dec(ascii=bb)
    >>> [Rass,Rm,Rchl,Rcyt,Rstarch,Rpyr,Rbio,Rphloem,Vstarch,ass13,disc,Rnew_starch,Rnew_cyt] = cuntz_gleixner(
    ...                jadecdate[1:], gpp, Rd, CO2air, Ra, gtot, jndecdate, Vcyt=Vcyt,
    ...                date0=jadecdate[0], V0starch=V0starch, R0starch=R0starch, R0cyt=R0cyt,
    ...                daynight=daynight,
    ...                daylength=daylength, Phi=Phi, s_resid=s_resid,
    ...                betas=betas, betap=betap, epsa=epsa, epsb=epsb,
    ...                epsg=epsg, epst=epst, epss=epss, epsp=epsp,
    ...                steady=False, fullmodel=True, julian=True)
    >>> # There are slight differences due to precision of dates
    >>> print(astr(Rphloem[0:6],5,pp=True))
    ['0.01092' '0.01091' '0.01089' '0.01087' '0.01085' '0.01085']
    >>> [Rass,Rm,Rchl,Rcyt,Rstarch,Rpyr,Rbio,Rphloem,Vstarch,ass13,disc,Rnew_starch,Rnew_cyt] = cuntz_gleixner(
    ...                jadecdate[1:], gpp, Rd, CO2air, Ra, gtot, jndecdate, Vcyt=Vcyt,
    ...                date0=jadecdate[0], V0starch=V0starch, R0starch=R0starch, R0cyt=R0cyt,
    ...                daynight=daynight,
    ...                daylength=daylength, Phi=Phi, s_resid=s_resid,
    ...                betas=betas, betap=2./3.-0.1, epsa=epsa, epsb=epsb,
    ...                epsg=epsg, epst=epst, epss=epss, epsp=epsp,
    ...                steady=False, fullmodel=True, julian=True, nocheck=True)
    >>> print(astr(Rphloem[0:6],5,pp=True))
    ['0.01092' '0.00956' '0.00932' '0.00906' '0.00889' '0.00909']


    License
    -------
    This file is part of the JAMS Python package, distributed under the MIT
    License. The JAMS Python package originates from the former UFZ Python
    library, Department of Computational Hydrosystems, Helmholtz Centre for
    Environmental Research - UFZ, Leipzig, Germany.

    Copyright (c) 2012-2021 Matthias Cuntz - mc (at) macu (dot) de

    Permission is hereby granted, free of charge, to any person obtaining a
    copy of this software and associated documentation files (the "Software"),
    to deal in the Software without restriction, including without limitation
    the rights to use, copy, modify, merge, publish, distribute, sublicense,
    and/or sell copies of the Software, and to permit persons to whom the
    Software is furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in
    all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.


    History
    -------
    Written,  Matthias Cuntz, Jan 2012
    Modified, Matthias Cuntz, Mar 2012 - julian
              Matthias Cuntz, May 2012 - nocheck
              Matthias Cuntz, May 2012 - Vcyt, daynight and betas=None default
              Matthias Cuntz, Feb 2013 - starch_mol2g, V0starchg
              Matthias Cuntz, Feb 2013 - ported to Python 3
              Matthias Cuntz, Nov 2016 - const.tiny -> const.eps
              Matthias Cuntz, Sep 2021 - code refactoring
    """
    #
    # Checks
    nss = False
    if not steady:
        nss = True
    nd = idecdate.size
    ng = iGPP.size
    nr = iRd.size
    nc = iCa.size
    na = iRa.size
    ns = igtot.size
    if ((nd != ng) | (nd != nr) | (nd != nc) | (nd != na) | (nd != ns)):
        raise ValueError('not all input sizes are equal')
    if (Rass | Rm | Rchl | Rcyt | Rstarch | Rpyr | Rbio | Rphloem |
        Vstarch | ass13 | disc | Rnew_starch | Rnew_cyt):
        fullmodel = False
    if fullmodel:
        Rass        = True
        Rm          = True
        Rchl        = True
        Rcyt        = True
        Rstarch     = True
        Rpyr        = True
        Rbio        = True
        Rphloem     = True
        Vstarch     = True
        ass13       = True
        disc        = True
        Rnew_starch = True
        Rnew_cyt    = True
    # Defaults
    # Day (1) or night (0)
    if not np.any(daynight is not None):
        daynight = np.where(iGPP > 0., 1, 0)
    isarr = np.ndim(daynight)
    if (isarr == 0):
        idaynight = np.ones(nd, dtype=int) * daynight
    else:
        idaynight = daynight
        nn = idaynight.size
        if (nn != nd):
            raise ValueError('daynight must be size 1 or size(idecdate)')
    # length of day [s]
    isarr = np.ndim(daylength)
    if (isarr == 0):
        idaylength = np.ones(nd) * daylength
    else:
        idaylength = daylength
        nn = idaylength.size
        if (nn != nd):
            raise ValueError('daylength must be size 1 or size(idecdate)')
    # Vo/Vc: ratio of carboxylateion to oxygentation of Rubisco
    isarr = np.ndim(Phi)
    if (isarr == 0):
        iPhi = np.ones(nd) * Phi
    else:
        iPhi = Phi
        nn = iPhi.size
        if (nn != nd):
            raise ValueError('Phi=Vo/Vc must be size 1 or size(idecdate)')
    # fraction of respiration occuring during biosynthesis: min=2/3; max=4/5
    isarr = np.ndim(betap)
    if (isarr == 0):
        ibetap = np.ones(nd) * betap
    else:
        ibetap = betap
        nn = ibetap.size
        if (nn != nd):
            raise ValueError('betap must be size 1 or size(idecdate)')
    mini = np.amin(ibetap)
    maxi = np.amax(ibetap)
    if not nocheck:
        if ((mini < 0.66) | (maxi > 0.8)):
            raise ValueError('betap must be: 2/3 < betap < 4/5')
    # betas the factor of leaf respiration that is transferred to biosynthesis
    # (default: 3*gpp/max(gpp))
    # betas*(1-betap) <= 1: if betap=2/3
    # -> betas<3: if betap=5/6 -> betas < 6
    if not np.any(betas is not None):
        betas = np.maximum((1./(1.-ibetap) * iGPP/np.amax(iGPP)) - const.eps,
                           0.)
    isarr = np.ndim(betas)
    if (isarr == 0):
        ibetas = np.ones(nd) * betas
    else:
        ibetas = betas
        nn = ibetas.size
        if (nn != nd):
            raise ValueError('betas must be size 1 or size(idecdate)')
    # effective fractionation through leaf boundary layer,
    # stomata and mesophyll
    isarr = np.ndim(epsa)
    if (isarr == 0):
        iepsa = np.ones(nd) * epsa
    else:
        iepsa = epsa
        nn = iepsa.size
        if (nn != nd):
            raise ValueError('epsa must be size 1 or size(idecdate)')
    # Vcyt
    if (nss & (~np.any(Vcyt is not None))):
        raise ValueError('Vcyt must be given if non-steady state')
    if nss:
        isarr = np.ndim(Vcyt)
        if (isarr == 0):
            iVcyt = np.ones(nd, dtype=float) * Vcyt
        else:
            iVcyt = Vcyt
            nn = iVcyt.size
            if (nn != nd):
                raise ValueError('Vcyt must be size 1 or size(idecdate)')
    # starch_mol2g
    if np.any(starch_mol2g is not None):
        isarr = np.ndim(starch_mol2g)
        if (isarr == 0):
            istarch_mol2g = np.ones(nd, dtype=float) * starch_mol2g
        else:
            istarch_mol2g = starch_mol2g
            nn = istarch_mol2g.size
            if (nn != nd):
                raise ValueError(
                    'starch_mol2g must be size 1 or size(idecdate)')
    else:
        istarch_mol2g  = np.ones(nd, dtype=float)
    #
    # derived variables
    iAss = np.abs(iGPP) - np.abs(iRd)  # Net assimilation
    iCc  = iCa - iAss/igtot            # CO2 @ chloroplast
    iVc  = np.abs(iGPP)/(1.-0.5*iPhi)  # carboxylation
    # seconds
    if julian:
        decfac = np.ones(nd)
    else:
        yr     = np.floor(idecdate)
        leap   = (((np.fmod(yr, 4) == 0) & (np.fmod(yr, 100) != 0)) |
                  (np.fmod(yr, 400) == 0))
        decfac = 365. + np.array(leap, dtype=float)
    isecs = np.rint(np.maximum(idecdate-np.roll(idecdate, 1), 0.) *
                    86400. * decfac)
    if np.size(date0) == 0:
        isecs[0] = np.rint((idecdate[1]-idecdate[0])*86400.*decfac[0])
    else:
        isecs[0] = np.rint((idecdate[0]-date0)*86400.*decfac[0])
    # seconds remainig until next sunrise
    nightlength     = 86400. - idaylength  # length of night before
    nightlength     = np.roll(nightlength, -1)
    nightlength[-1] = nightlength[-2]
    dsecs = idaylength + nightlength
    nsecs = np.empty(nd)
    issun = sunrise
    sr    = np.rint((idecdate - issun) * 86400. * decfac)
    for i in range(nd):
        if sr[i] > 0.:
            sr[i:] -= dsecs[i]
        nsecs[i] = -sr[i]
    nsecs = np.roll(nsecs, 1)
    if np.size(date0) == 0:
        nsecs[0] = nsecs[0] + (nsecs[0]-nsecs[1])
    else:
        sr0 = np.rint((date0-issun)*86400.*decfac[0])
        if sr0 > 0.:
            sr0 -= dsecs[0]
        nsecs[0] = -sr0
    #
    # fraction of starch synthesis,
    # same as T of Tchekez but with different units
    tmp    = 1./86400.
    ibigT  = 1. - idaylength*tmp  # =0.056*6 @ Tcherkez
    ibetar = 1. - ibetas*(1.-ibetap)
    if not nocheck:
        if np.any(ibetar <= 0.):
            raise ValueError('betar = 1-betas*(1-betap) <= 0.')
    # limit to eps=1000 permil
    iepsr  = np.where(ibetar > const.eps, epss * (1.-ibetar)/ibetar, 1.)
    iepsg1 = epsg*0.5*iPhi + ibigT*(1.-0.5*iPhi)*epst
    iepsp1 = (epss*ibetas + iepsr*ibetar) / (ibetas + ibetar)
    #
    # Define output
    iAss13       = np.empty(nd)      # 13CO2 flux
    iRass        = np.empty(nd)      # Input
    idisc        = np.empty(nd)      # Discrimination
    iRm          = np.empty(nd)      # at site of carboxylation
    iRchl        = np.empty(nd)      # sucrose in chloroplast
    iRstarch     = np.empty(nd)      # accumulated starch in chloroplast
    iVstarch     = np.empty(nd)      # Starch concentration
    iVstarchg    = np.empty(nd)      # Starch concentration in gC/gDW
    iRcyt        = np.empty(nd)      # sucrose in cytoplast
    iRpyr        = np.empty(nd)      # pyruvate in cytoplast
    iRbio        = np.empty(nd)      # biosynthesis in cytoplast
    iRnew_starch = np.empty(nd)      # new starch in chloroplast
    iRnew_cyt    = np.empty(nd)      # new sucrose in cytoplasm
    iRphloem     = np.empty(nd)      # new phloem
    #
    if nss:
        dVcytdt = (iVcyt-np.roll(iVcyt, 1))/isecs
        dVcytdt[0] = dVcytdt[1]
    #
    # Calc model
    if nss:  # non-steady-state
        for i in range(nd):
            if idaynight[i] == 1:  # nss day
                tmp1     = (1.-ibigT[i]) * (1.-0.5*iPhi[i]) * iVc[i]
                tmp2     = (1.-epsb) * iVc[i] * (1.-(1.-epsg)/(1.-iepsg1[i]) *
                                                 0.5*iPhi[i])
                tmp3     = (1.-epsb) / (1.-iepsg1[i])
                zaehler  = (1.-iepsa[i]) * igtot[i] * iCa[i] * iRa[i]
                nenner   = tmp2 + (1.-iepsa[i]) * igtot[i] * iCc[i]
                k1       = tmp1 * tmp3 * zaehler/nenner / iVcyt[i]
                k2       = ( (1. - tmp3*iRd[i]/(1.-iepsp1[i])/nenner) * tmp1 +
                             dVcytdt[i] ) / iVcyt[i]
                Rcytm1   = R0cyt if i == 0 else iRcyt[i-1]
                iRcyt[i]        = k1/k2 + (Rcytm1-k1/k2) * np.exp(-k2*isecs[i])
                iRm[i]          = zaehler/nenner + (iRd[i] / (1.-iepsp1[i]) /
                                                    nenner * iRcyt[i])
                iRchl[i]        = (1.-epsb) / (1.-iepsg1[i])*iRm[i]
                iRpyr[i]        = 1./(1.-iepsp1[i]) * iRcyt[i]
                iRbio[i]        = (1.-epss) / (1.-iepsp1[i]) * iRcyt[i]
                iRnew_starch[i] = (1.-epst) * iRchl[i]
                iRnew_cyt[i]    = iRchl[i]
                fphloem1        = (1.-ibigT[i]) * np.abs(iGPP[i]) - (
                    (ibetar[i] + ibetas[i]) * iRd[i])
                iRphloem[i]     = fphloem1 * iRcyt[i]
                fphloem2        = ibetap[i] * ibetas[i] * iRd[i]
                iRphloem[i]    += fphloem2 * (1.-epsp) * iRbio[i]
                iRphloem[i]    /= fphloem1 + fphloem2
                # Integrate starch
                fstarch      = ibigT[i] * np.abs(iGPP[i])
                Vstarchm1    = V0starch  if i == 0 else iVstarch[i-1]
                Vstarchgm1   = V0starchg if i == 0 else iVstarchg[i-1]
                Rstarchm1    = R0starch  if i == 0 else iRstarch[i-1]
                iVstarch[i]  = Vstarchm1  + fstarch*isecs[i]
                iVstarchg[i] = Vstarchgm1 + fstarch*isecs[i]*istarch_mol2g[i]
                iRstarch[i]  = (Rstarchm1*Vstarchm1 +
                                fstarch*isecs[i]*iRnew_starch[i]) / iVstarch[i]
            else:  # nss night
                Vstarchm1       = V0starch  if i == 0 else iVstarch[i-1]
                Vstarchgm1      = V0starchg if i == 0 else iVstarchg[i-1]
                fstarch         = -(Vstarchm1-s_resid)/nsecs[i]
                iVstarch[i]     = Vstarchm1  + fstarch*isecs[i]
                iVstarchg[i]    = Vstarchgm1 + (fstarch * isecs[i] *
                                                istarch_mol2g[i])
                Rstarchm1       = R0starch  if i == 0 else iRstarch[i-1]
                iRstarch[i]     = Rstarchm1
                iRnew_starch[i] = Rstarchm1
                iRchl[i]        = iRnew_starch[i]
                fs              = np.abs(fstarch)
                k1              = fs/iVcyt[i]*iRchl[i]
                k2              = (fs+np.abs(dVcytdt[i])) / iVcyt[i]
                Rcytm1          = R0cyt if i == 0 else iRcyt[i-1]
                iRcyt[i]        = k1/k2 + (Rcytm1-k1/k2) * np.exp(-k2*isecs[i])
                iRpyr[i]        = iRcyt[i]
                iRbio[i]        = iRpyr[i] if i == 0 else iRbio[i-1]
                iRm[i]          = iCa[i]/iCc[i]*iRa[i] + ((1.-iCa[i]/iCc[i]) *
                                                          iRpyr[i] /
                                                          (1.-iepsa[i]))
                iRnew_cyt[i]    = iRchl[i]
                iRphloem[i]     = iRcyt[i]
    else:  # steady-state
        for i in range(nd):
            if idaynight[i] == 1:  # ss day
                tmp1 = (1.-iepsa[i]) * igtot[i] * iRa[i] * iCa[i]
                tmp2 = (1.-iepsa[i]) * igtot[i] * iCc[i]
                tmp3 = ( (1.-epsb) * iVc[i] * (1. - (1.-epsg) /
                                               (1.-iepsg1[i])*0.5*iPhi[i]) -
                         (1.-epsb) / ((1.-iepsp1[i])*(1.-iepsg1[i])) * iRd[i] )
                iRm[i]          = tmp1 / (tmp2+tmp3)
                iRchl[i]        = (1.-epsb) / (1.-iepsg1[i]) * iRm[i]
                iRcyt[i]        = (1.-epsb) / (1.-iepsg1[i]) * iRm[i]
                iRpyr[i]        = ((1.-epsb) / (1.-iepsg1[i]) /
                                   (1.-iepsp1[i]) * iRm[i])
                iRbio[i]        = ((1.-epsb) / (1.-iepsg1[i]) /
                                   (1.-iepsp1[i]) * (1.-epss) * iRm[i])
                iRnew_starch[i] = (1.-epst) * iRchl[i]
                iRnew_cyt[i]    = iRchl[i]
                fphloem1        = ((1.-ibigT[i]) * np.abs(iGPP[i]) -
                                   (ibetar[i]+ibetas[i]) * iRd[i])
                iRphloem[i]     = fphloem1 * iRcyt[i]
                fphloem2        = ibetap[i] * ibetas[i] * iRd[i]
                iRphloem[i]    += fphloem2 * (1.-epsp) * iRbio[i]
                iRphloem[i]    /= fphloem1 + fphloem2
                # Integrate starch
                fstarch      = ibigT[i] * np.abs(iGPP[i])
                Vstarchm1    = V0starch  if i == 0 else iVstarch[i-1]
                Vstarchgm1   = V0starchg if i == 0 else iVstarchg[i-1]
                Rstarchm1    = R0starch  if i == 0 else iRstarch[i-1]
                iVstarch[i]  = Vstarchm1  + fstarch*isecs[i]
                iVstarchg[i] = Vstarchgm1 + fstarch*isecs[i]*istarch_mol2g[i]
                iRstarch[i]  = (Rstarchm1*Vstarchm1 +
                                fstarch*isecs[i]*iRnew_starch[i]) / iVstarch[i]
            else:  # ss night
                Vstarchm1       = V0starch  if i == 0 else iVstarch[i-1]
                Vstarchgm1      = V0starchg if i == 0 else iVstarchg[i-1]
                fstarch         = -(Vstarchm1-s_resid)/nsecs[i]
                iVstarch[i]     = Vstarchm1  + fstarch*isecs[i]
                iVstarchg[i]    = (Vstarchgm1 +
                                   fstarch*isecs[i]*istarch_mol2g[i])
                Rstarchm1       = R0starch  if i == 0 else iRstarch[i-1]
                iRstarch[i]     = Rstarchm1
                iRnew_starch[i] = Rstarchm1
                iRchl[i]        = Rstarchm1
                iRcyt[i]        = iRchl[i]
                iRpyr[i]        = iRcyt[i]
                iRbio[i]        = iRpyr[i] if i == 0 else iRbio[i-1]
                iRm[i]          = (iCa[i]/iCc[i]*iRa[i] +
                                   (1.-iCa[i]/iCc[i])*iRpyr[i]/(1.-iepsa[i]))
                iRnew_cyt[i]    = iRchl[i]
                iRphloem[i]     = iRcyt[i]
    #
    iAss13 = (1.-iepsa) * igtot * (iRa*iCa-iRm*iCc)
    iRass  = iAss13 / iAss
    idisc  = 1. - iRass / iRa
    #
    out  = []
    if Rass:
        out += [iRass]
    if Rm:
        out += [iRm]
    if Rchl:
        out += [iRchl]
    if Rcyt:
        out += [iRcyt]
    if Rstarch:
        out += [iRstarch]
    if Rpyr:
        out += [iRpyr]
    if Rbio:
        out += [iRbio]
    if Rphloem:
        out += [iRphloem]
    if Vstarch:
        out += [iVstarch]
    if ass13:
        out += [iAss13]
    if disc:
        out += [idisc]
    if Rnew_starch:
        out += [iRnew_starch]
    if Rnew_cyt:
        out += [iRnew_cyt]
    if Vstarch & np.any(starch_mol2g is not None):
        out += [iVstarchg]

    return out


if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)



#!/usr/bin/env python
import numpy as np
import xlrd


__all__ = ['get_header_excel', 'get_value_excel']


# --------------------------------------------------------------------

def get_header_excel(excelfile, sheet):
    """
        Get the header row of an Excel sheet.

        Same as get_value_excel(excelfile, sheet, '', None).


        Input
        -----
        excelfile   Filename of Excel file
        sheet       Name of Sheet in Excel file


        Output
        ------
        list with entries of first row of Excel sheet


        Examples
        --------
        --> see __init__.py for full example of workflow

        hh = get_header_excel('CHS-measurements.xlsx', 'Logger Hohes Holz')
        -> ['variable description', 'date start', 'date end', ... ]


        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT
        License. The JAMS Python package originates from the former UFZ Python library,
        Department of Computational Hydrosystems, Helmholtz Centre for Environmental
        Research - UFZ, Leipzig, Germany.

        Copyright (c) 2016 Matthias Cuntz - mc (at) macu (dot) de

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  MC, Mar 2016
    """
    # open Excel file
    wb = xlrd.open_workbook(excelfile)
    # check sheet exists
    if sheet not in wb.sheet_names(): raise ValueError('No sheet '+sheet+' in file '+excelfile)
    sh = wb.sheet_by_name(sheet)
    # Return variable and column names
    columns   = sh.row_values(0,0)

    # relase Excel file
    del wb

    return columns


# --------------------------------------------------------------------

def get_value_excel(excelfile, sheet, variable, column, all_rows=False):

    """
        Get a cell in a sheet of an Excel file
        where the column is identified by the name in the first row
        and the row is identified by the value in the column "headerout (final)".

        Both, variable and column can be iterables.

        If variable and/or column is None then the list of variables and columns is returned.


        Input
        -----
        excelfile   Filename of Excel file
        sheet       Name of Sheet in Excel file
        variable    Variable name in column "headerout (final)"
        column      Name in header line


        Optional Input
        --------------
        all_rows    option to return whole column, i.e. all rows (default: False).
                    Should be chosen if there are/may be duplicates in vars/the "headerout (final)" column;
                    the result of get_value_excel(file,sheet,vars,col)
                    if vars = get_value_excel(file,sheet,None,'') is NOT equal to
                    get_value_excel(file,sheet,None,col,all_rows=True)
                    as the former cannot deal properly with duplicates in vars.


        Output
        ------
        content of the desired cell
        if column or variable is iterable, output is list
        if column and variable is iterable, output is list (variable) of list (columns)
        if column and/or variable is not given, then a list of all variable and/or columns is returned
        if column is iterable and all_rows==True, output is list of list (all rows of all columns)


        Examples
        --------
        --> see __init__.py for full example of workflow

        mini = get_value_excel('CHS-measurements.xlsx', 'Logger Hohes Holz', 'BC1_Ptemp [degC]', 'Min')
        -> -20.0
        mini, maxi = get_value_excel('CHS-measurements.xlsx', 'Logger Hohes Holz', 'BC1_Ptemp [degC]', ['Min', 'Max'])
        -> -20.0, 50.0
        min1, min2 = get_value_excel('CHS-measurements.xlsx', 'Logger Hohes Holz', ['BC1_Ptemp [degC]', 'Tair50HMP [degC]'], 'Min')
        -> -20.0, -20.0
        minmax1, minmax2 = get_value_excel('CHS-measurements.xlsx', 'Logger Hohes Holz', ['BC1_Ptemp [degC]', 'Tair50HMP [degC]'], ['Min', 'Max'])
        -> [-20.0, 50.0], [-20.0, 50.0]
        vars = get_value_excel('CHS-measurements.xlsx', 'Logger Hohes Holz', None, '')
        columns = get_value_excel('CHS-measurements.xlsx', 'Logger Hohes Holz', '', None)
        vars, columns = get_value_excel('CHS-measurements.xlsx', 'Logger Hohes Holz', None, None)


        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT
        License. The JAMS Python package originates from the former UFZ Python library,
        Department of Computational Hydrosystems, Helmholtz Centre for Environmental
        Research - UFZ, Leipzig, Germany.

        Copyright (c) 2015-2016 Matthias Cuntz, Benjamin Dechant - mc (at) macu (dot) de

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  MC, Mar 2015
        Modified, MC, Jul 2015 - column/variable can be iterable
                               - column/variable can None, returning names
                      Aug 2015 - added docu of input variables
                  BD, Nov 2016 - all_rows option to get the whole column (i.e. all rows)
                                 in cases where the varnames contain duplicates
    """
    # open Excel file
    wb = xlrd.open_workbook(excelfile)
    # check sheet exists
    if sheet not in wb.sheet_names(): raise ValueError('No sheet '+sheet+' in file '+excelfile)
    sh = wb.sheet_by_name(sheet)
    # Return variable and column names
    columns   = sh.row_values(0,0)
    variables = sh.col_values(columns.index('headerout (final)'), start_rowx=1, end_rowx=sh.nrows)
    if (variable is None) and (column is None):
        return [variables, columns]
    elif (variable is None) and (not all_rows):
        return variables
    elif (column is None):
        return columns
    # check columns exist
    hh = ['headerout (final)']
    if isinstance(column, (list, tuple, np.ndarray)):
        hh.extend(column)
    else:
        hh.append(column)
    for i in hh:
        if i not in columns:
            raise ValueError('Column '+i+' not found in sheet '+sheet+' of file '+excelfile)
    # check columns
    if isinstance(variable, (list, tuple, np.ndarray)) and (not all_rows):
        for ivar in variable:
            if ivar not in variables:
                raise ValueError(ivar+' not in column "headerout (final)" in sheet '+sheet+' of file '+excelfile)
    elif not all_rows:
        if variable not in variables:
            raise ValueError(variable+' not in column "headerout (final)" in sheet '+sheet+' of file '+excelfile)
    # extract cells
    if isinstance(column, (list, tuple, np.ndarray)) and isinstance(variable, (list, tuple, np.ndarray)) and (not all_rows):
        out = list()
        for ivar in variable:
            out.append([ sh.col_values(columns.index(icol),
                                       start_rowx=1, end_rowx=sh.nrows)[variables.index(ivar)] for icol in column ])
    elif all_rows: # gives all rows for the column if no variable is given and all_rows==True
        out=[ sh.col_values(columns.index(icol),
                        start_rowx=1, end_rowx=sh.nrows) for icol in column ]

    else:
        if isinstance(column, (list, tuple, np.ndarray)):     # columns
            out = [ sh.col_values(columns.index(icol),
                                  start_rowx=1, end_rowx=sh.nrows)[variables.index(variable)] for icol in column ]
        elif isinstance(variable, (list, tuple, np.ndarray)): # variables
            out = [ sh.col_values(columns.index(column),
                                  start_rowx=1, end_rowx=sh.nrows)[variables.index(ivar)] for ivar in variable ]
        else:                                                 # scalars
            parameter = sh.col_values(columns.index(column), start_rowx=1, end_rowx=sh.nrows)
            out = parameter[variables.index(variable)]
    # relase Excel file
    del wb

    return out


# --------------------------------------------------------------------

if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)

    # base_folder = '/Volumes/Gruppen/CHS-Data'
    # setup_chs = base_folder+'/CHS-measurements.xlsx'
    # sheet = 'Logger Hohes Holz'
    # v = 'Tair50HMP [degC]'
    # print(get_value_excel(setup_chs, sheet, v, 'Min'))
    # print(get_value_excel(setup_chs, sheet, v, ['Min','Max']))
    # v = ['Tair50HMP [degC]', 'BC1_Ptemp [degC]']
    # print(get_value_excel(setup_chs, sheet, v, 'Min'))
    # print(get_value_excel(setup_chs, sheet, v, ['Min','Max']))
    # # print(get_value_excel(setup_chs, 'Logger Hohes Holz', None, ''))
    # print(get_value_excel(setup_chs, 'Logger Hohes Holz', '', None))
    # # print(get_value_excel(setup_chs, 'Logger Hohes Holz', None, None))
    # print(get_value_excel(setup_chs, 'Logger Hohes Holz', None, ''))

    # print(get_header_excel(setup_chs, sheet))

    # v = 'BC1_Voltage [V]'
    # print(get_value_excel(setup_chs, sheet, v, ['Flag_0', 'Flag_1', 'Flag_2', 'Flag_3', 'Flag_4', 'Flag_5']))



#!/usr/bin/env python
import numpy as np
from pyjams.jams.correlate import correlate


def lagcorr(y1, y2, max=True, plot=False):
    """
        Calculates the time lag between two 1D arrays where the correlation
        of both arrays is maximum or minimum.


        Definition
        ----------
        def lagcorr(y1, y2, max=True, plot=False):


        Input
        -----
        y1           np.ma.array(N) or unmasked array of size N
        y2           np.ma.array(N) or unmasked array of size N


        Optional Input
        --------------
        max          bool: if True, returns lag of maximum correlation, else
                           returns lag of minimum correlation
        plot         bool: if True, correlation function is plotted


        Output
        ------
        lag          int: lag of maximum or minimum correlation between y1 and y2


        Examples
        --------
        >>> # Create some data
        >>> y1 = np.sin(np.arange(np.pi,np.pi*3,0.1))[2:]
        >>> y2 = np.sin(np.arange(np.pi,np.pi*3,0.1))[:-2]
        >>> # calculate correlation lag
        >>> lagcorr(y1, y2)
        2

        >>> x = np.sin(np.random.random(1000)) # some data
        >>> y = np.roll(x, -100)                 # shift by 100 to the right
        >>> lagcorr(x, y)
        -100
        >>> y = np.roll(x, 100)                 # shift by 100 to the left
        >>> lagcorr(x, y)
        100


        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT
        License. The JAMS Python package originates from the former UFZ Python library,
        Department of Computational Hydrosystems, Helmholtz Centre for Environmental
        Research - UFZ, Leipzig, Germany.

        Copyright (c) 2014 Arndt Piayda, Matthias Cuntz - mc (at) macu (dot) de

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  AP, Mar 2014
        Modified, AP & MC, Apr 2014 - use correlate
    """

    # cross correlation with Fast Fourier
    cc = correlate(y1,y2)
    # minimum or maximum
    if max:
        out = np.argmax(cc)
    else:
        out = np.argmin(cc)
    # if negative lag
    if out > y1.size: out = out - 2*y1.size

    # plot
    if plot:
        import matplotlib.pyplot as plt
        fig = plt.figure('maxcorr')
        sub = fig.add_subplot(111)
        xx = np.arange(cc.size)
        sub.plot(xx,cc,'-')
        sub.plot(xx[out],cc[out],'o')
        plt.show()

    return out

if __name__ == '__main__':
    import doctest
    doctest.testmod()

    # # Create some data
    # y1 = np.sin(np.arange(np.pi,np.pi*3,0.1))[2:]
    # y2 = np.sin(np.arange(np.pi,np.pi*3,0.1))[:-2]
    # # calculate correlation lag
    # lagcorr(y1, y2)



#!/usr/bin/env python
from __future__ import division, absolute_import, print_function
"""
Created on Sun Aug 30 18:35:32 2015

@author: wiedeman
"""
import numpy as np

"""
        Checks if abs. differences of array values within a certain window are smaller than threshold.
        Return mask. 


        Definition
        ----------
        def samevalue(x,tol,window)


        Input
        -----
        x            1D-array
        tol          tolerance for difference of two subsequent values.
                     If absolute difference between two values is smaller than tol array entries will be masked
        window       size of window where all abs. differences of subsequent values has to be smaller than tol to get masked
                     if at least one difference is larger than tol all array entries within window stay unmasked
        

        Output
        ------
        new_mask     mask of x where abs. difference of two subsequent values within a certain window are smaller that tol


        Examples
        --------
        >>> # Create some data
        >>> x                = np.arange(0.,10.,0.01)
        >>> x_curve          = (np.sin(x))
        >>> x_curve[700:800] = x_curve[699]
        >>> x_mask           = samevalue(x_curve,0.001,50)
        >>> xx               = ma.array(x_curve, mask=x_mask)

        >>> # apply new mask to x
        >>> print(x_mask[650:850])
        [False, False, False, False, False, False, False, False, False,
       False, False, False, False, False, False, False, False, False,
       False, False, False, False, False, False, False, False, False,
       False, False, False, False, False, False, False, False, False,
       False, False, False, False, False, False, False, False, False,
       False, False, False, False,  True,  True,  True,  True,  True,
        True,  True,  True,  True,  True,  True,  True,  True,  True,
        True,  True,  True,  True,  True,  True,  True,  True,  True,
        True,  True,  True,  True,  True,  True,  True,  True,  True,
        True,  True,  True,  True,  True,  True,  True,  True,  True,
        True,  True,  True,  True,  True,  True,  True,  True,  True,
        True,  True,  True,  True,  True,  True,  True,  True,  True,
        True,  True,  True,  True,  True,  True,  True,  True,  True,
        True,  True,  True,  True,  True,  True,  True,  True,  True,
        True,  True,  True,  True,  True,  True,  True,  True,  True,
        True,  True,  True,  True,  True,  True,  True,  True,  True,
        True,  True,  True,  True,  True,  True, False, False, False,
       False, False, False, False, False, False, False, False, False,
       False, False, False, False, False, False, False, False, False,
       False, False, False, False, False, False, False, False, False,
       False, False, False, False, False, False, False, False, False,
       False, False, False, False, False, False, False, False, False,
       False, False]


        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT
        License. The JAMS Python package originates from the former UFZ Python library,
        Department of Computational Hydrosystems, Helmholtz Centre for Environmental
        Research - UFZ, Leipzig, Germany.

        Copyright (c) 2015 Andreas Wiedemann

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  AW, Aug 2015

"""
def samevalue(x,tol,window):
    # define mask where input values smaller than tolerance
    x         = np.ma.array(x)        
    diff      = np.abs(np.diff(x))
    tol       = np.float(tol)
    x_mask    = np.ma.masked_where(diff > tol, diff)
    
    # mask x only if values are smaller than tolerance in a range as long or longer than the window-size  
    count = 0    
    for i in range(np.shape(x)[0]):
        if i > window and sum(x_mask.mask[count:][i-window:i]) < 1:
            first_unique_index                      = i-window+1
            last_unique_index                       = np.argmax(diff[first_unique_index:] > tol)
            x[first_unique_index:first_unique_index + last_unique_index] = np.ma.masked
            count += last_unique_index
        else:
            count += 1
        
    return x.mask

if __name__ == '__main__':
    import doctest
    doctest.testmod()



#!/usr/bin/env python
from __future__ import division, absolute_import, print_function
import numpy as np
from scipy.spatial.distance import cdist

def get_nearest(xy, xyz):
    """
        Returns the z value for each point in an xy coordinate array 
        which is equal to the z value at the nearest point in a given
        xyz field.

        Definition
        ----------
        get_nearest(xy, xyz):


        Input
        -----
        xy        2D np.array (n,2),
                  x and y coordinates of the points where a z value is desired
        xyz       2D np.array (m,3), x, y and z vaules of the field


        Output
        ------
        z         1D np.array (n,), z values at all points of xy


        Restrictions
        ------------
        Does not work with NAN values or masked arrays. You need to provide only
        valid data.


        Example
        --------
        # make some data
        >>> np.random.seed(seed=1)
        >>> xy = np.random.random(6).reshape(-1,2)
        >>> np.random.seed(seed=2)
        >>> xyz = np.random.random(150).reshape(-1,3)

        # get z values at the points xy
        >>> z = get_nearest(xy, xyz)
        >>> print(np.round(z,2))
        [0.25 0.7  0.28]


        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT
        License. The JAMS Python package originates from the former UFZ Python library,
        Department of Computational Hydrosystems, Helmholtz Centre for Environmental
        Research - UFZ, Leipzig, Germany.

        Copyright (c) 2014 Arndt Piayda

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  AP, May 2014
        Modified, ST, Jun 2014 - minor change to documentation
    """

    dist   = cdist(xy, xyz[:,:2], metric='euclidean')
    dist   = np.ma.array(dist, mask=dist==0.)
    indmin = np.ma.argmin(dist, axis=1)
    
    return xyz[indmin,2]


if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)



"""
Provides a variety of special functions, including common test functions for
parameter estimations such as Rosenbrock and Griewank, test functions for
parameter sensitivity analysis such as the Ishigami and Homma function, several
forms of the logistic function and its first and second derivatives, and a
variety of other functions together with robust and square cost functions to
use with the ``scipy.optimize`` package.

:copyright: Copyright 2014-2021 Matthias Cuntz, see AUTHORS.rst for details.
:license: MIT License, see LICENSE for details.

Subpackages
===========
.. autosummary::
   general_functions
   fit_functions
   logistic_function
   opti_test_functions
   sa_test_functions
"""
from .general_functions import curvature
from .fit_functions import cost_abs, cost_square
from .fit_functions import arrhenius, arrhenius_p
from .fit_functions import cost_arrhenius, cost2_arrhenius
from .fit_functions import f1x, f1x_p, cost_f1x, cost2_f1x
from .fit_functions import fexp, fexp_p, cost_fexp, cost2_fexp
from .fit_functions import gauss, gauss_p, cost_gauss, cost2_gauss
from .fit_functions import lasslop, lasslop_p, cost_lasslop, cost2_lasslop
from .fit_functions import line, line_p, cost_line, cost2_line
from .fit_functions import line0, line0_p, cost_line0, cost2_line0
from .fit_functions import lloyd_fix, lloyd_fix_p
from .fit_functions import cost_lloyd_fix, cost2_lloyd_fix
from .fit_functions import lloyd_only_rref, lloyd_only_rref_p
from .fit_functions import cost_lloyd_only_rref, cost2_lloyd_only_rref
from .fit_functions import sabx, sabx_p, cost_sabx, cost2_sabx
from .fit_functions import poly, poly_p, cost_poly, cost2_poly
from .fit_functions import cost_logistic, cost2_logistic
from .fit_functions import cost_logistic_offset, cost2_logistic_offset
from .fit_functions import cost_logistic2_offset, cost2_logistic2_offset
from .fit_functions import see, see_p, cost_see, cost2_see
from .logistic_function import logistic, logistic_p
from .logistic_function import dlogistic, dlogistic_p
from .logistic_function import d2logistic, d2logistic_p
from .logistic_function import logistic_offset, logistic_offset_p
from .logistic_function import dlogistic_offset, dlogistic_offset_p
from .logistic_function import d2logistic_offset, d2logistic_offset_p
from .logistic_function import logistic2_offset, logistic2_offset_p
from .logistic_function import dlogistic2_offset, dlogistic2_offset_p
from .logistic_function import d2logistic2_offset, d2logistic2_offset_p
from .opti_test_functions import ackley, griewank, goldstein_price, rastrigin
from .opti_test_functions import rosenbrock, six_hump_camelback
from .sa_test_functions import B, g, G, Gstar, K, bratley, fmorris, morris
from .sa_test_functions import oakley_ohagan, ishigami_homma
from .sa_test_functions import linear, product, ratio, ishigami_homma_easy

__all__ = ['curvature']
__all__ += ['cost_abs', 'cost_square',
            'arrhenius', 'arrhenius_p', 'cost_arrhenius', 'cost2_arrhenius',
            'f1x', 'f1x_p', 'cost_f1x', 'cost2_f1x',
            'fexp', 'fexp_p', 'cost_fexp', 'cost2_fexp',
            'gauss', 'gauss_p', 'cost_gauss', 'cost2_gauss',
            'lasslop', 'lasslop_p', 'cost_lasslop', 'cost2_lasslop',
            'line', 'line_p', 'cost_line', 'cost2_line',
            'line0', 'line0_p', 'cost_line0', 'cost2_line0',
            'lloyd_fix', 'lloyd_fix_p', 'cost_lloyd_fix', 'cost2_lloyd_fix',
            'lloyd_only_rref', 'lloyd_only_rref_p', 'cost_lloyd_only_rref',
            'cost2_lloyd_only_rref',
            'sabx', 'sabx_p', 'cost_sabx', 'cost2_sabx',
            'poly', 'poly_p', 'cost_poly', 'cost2_poly',
            'cost_logistic', 'cost2_logistic',
            'cost_logistic_offset', 'cost2_logistic_offset',
            'cost_logistic2_offset', 'cost2_logistic2_offset',
            'see', 'see_p', 'cost_see', 'cost2_see']
__all__ += ['logistic', 'logistic_p',
            'dlogistic', 'dlogistic_p',
            'd2logistic', 'd2logistic_p',
            'logistic_offset', 'logistic_offset_p',
            'dlogistic_offset', 'dlogistic_offset_p',
            'd2logistic_offset', 'd2logistic_offset_p',
            'logistic2_offset', 'logistic2_offset_p',
            'dlogistic2_offset', 'dlogistic2_offset_p',
            'd2logistic2_offset', 'd2logistic2_offset_p']
__all__ += ['ackley', 'griewank', 'goldstein_price', 'rastrigin',
            'rosenbrock', 'six_hump_camelback']
__all__ += ['B', 'g', 'G', 'Gstar', 'K', 'bratley', 'fmorris', 'morris',
            'oakley_ohagan', 'ishigami_homma',
            'linear', 'product', 'ratio', 'ishigami_homma_easy']



#!/usr/bin/env python
import numpy as np
from pyjams.jams.date2dec import date2dec


__all__ = ['t2sap']


def t2sap(date, data, swd=None, undef=-9999.):
    """
    Conversion of temperature difference measured with sap flow sensors
    (Granier type) in (mV) to sap flux density (cm^3 cm^-2 h^-1) (Granier,
    1987).

    In addition, the correction according to Clearwater (1999) is possible.


    Definition
    ----------
    def t2sap(date, data, swd=None, undef=-9999.):


    Input
    -----
    date    1D array (n,) of ascii times in format DD.MM.YYYY hh:mm:ss
    data    ND array (n,...) of the raw data in mV. First dimension is time


    Optional Input
    --------------
    undef   values are excluded from calculations (default: -9999)
    swd     if None:     sapflux calculation according to original Granier
                         calibration function.
            if not None: sapwood depth in cm for Clearwater correction
                         T = (T - b*Tmax)/a
                         with a = active sapwood   = swd/2
                              b = inactive sapwood = 1-a


    Output
    ------
    SFD     2D array (n,m) with sap flux density in [cm^3 cm^-2 h^-1] according
            to the original Granier calibration function:
            SFD = 0.0119*K**1.231*3600 [Granier, 1985]


    References
    ----------
    Granier, A.
        Evaluation of transpiration in a Douglas-fir stand by means of
        sap flow measurements, Tree Physiology 3, 309-320, 1987
    Clearwater, M. J., Meinzer, F. C., Andrade, J. L., Goldstein, G.,
        Holbrook, N. M., Potential errors in measurement of nonuniform
        sap flow using heat dissipation probes, Tree Physiology 19, 681-687,
        1999


    Examples
    --------
    # normal sapflux conversion
    >>> data = np.array([0.434, 0.433, 0.432, 0.431, 0.431, 0.432])
    >>> date = ['18.05.2013 08:00', '18.05.2013 08:10', '18.05.2013 08:20',
    ...         '18.05.2013 08:30', '18.05.2013 08:40', '18.05.2013 08:50']
    >>> SFD = t2sap(date, data)
    >>> print(np.round(SFD,3))
    [0.    0.024 0.057 0.095 0.095 0.057]


    >>> # sapflux conversion including clearwater correction
    >>> data = np.array([0.434, 0.433, 0.432, 0.431, 0.431, 0.432])
    >>> date = ['18.05.2013 08:00', '18.05.2013 08:10', '18.05.2013 08:20',
    ...         '18.05.2013 08:30', '18.05.2013 08:40', '18.05.2013 08:50']
    >>> SFD  = t2sap(date, data, swd=1.5)
    >>> print(np.round(SFD,3))
    [0.    0.035 0.082 0.135 0.135 0.082]


    License
    -------
    This file is part of the JAMS Python package, distributed under the MIT
    License. The JAMS Python package originates from the former UFZ Python
    library, Department of Computational Hydrosystems, Helmholtz Centre for
    Environmental Research - UFZ, Leipzig, Germany.

    Copyright (c) 2014-2021 Andreas Wiedemann, Matthias Cuntz - mc (at) macu
    (dot) de

    Permission is hereby granted, free of charge, to any person obtaining a
    copy of this software and associated documentation files (the "Software"),
    to deal in the Software without restriction, including without limitation
    the rights to use, copy, modify, merge, publish, distribute, sublicense,
    and/or sell copies of the Software, and to permit persons to whom the
    Software is furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in
    all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.


    History
    -------
    Written,  Andreas Wiedemann, Jun 2014
    Modified, Matthias Cuntz, Jun 2014
                  - sap_app -> t2sap, incl. undef, first date then data
                  - squeeze 1D array, fill undef if not masked
              Matthias Cuntz, Nov 2016 - ported to Python 3
              Matthias Cuntz, Sep 2021 - code refactoring
    """
    isvec = False
    if data.ndim == 1:
        isvec = True
        data = data.reshape((-1, 1))
    # mask data
    isnotmask = True
    if type(data) == np.ma.core.MaskedArray:
        isnotmask = False
    data_masked = np.ma.array(data, mask=(data == undef))

    # julian day
    jd        = np.array(date2dec(ascii=date))
    # integer julian day
    jd_int    = jd.astype(int)
    # unique days
    jd_uni    = np.unique(jd_int)
    # Tmax per day
    tmax_day  = np.ma.ones((jd_uni.size, data.shape[1]))*undef
    # Time of Tmax per day
    jdmax_day = np.ma.zeros((jd_uni.size, data.shape[1]), dtype=int)

    # Determine Tmax per day
    for count, i in enumerate(jd_uni):
        # where is given day
        ii = np.where(jd_int == i)[0]
        # index of Tmax of day
        jj = np.ma.argmax(data_masked[ii, :], axis=0)
        # time at Tmax of day
        jdmax_day[count, :] = jd[ii[jj]]
        # Tmax per day
        tmax_day[count, :] = data_masked[ii[jj], :]

    # Tmax for every record
    tmax = np.empty(data.shape)
    for i in range(data.shape[1]):
        # time points at Tmax per day
        xx = jdmax_day[:, i]
        # Tmax per day
        yy = tmax_day[:, i]
        ii = ~yy.mask
        # Tmax per time point
        tmax[:, i] = np.interp(jd, xx[ii], yy[ii])

    if swd is None:
        Tsw = data_masked
    else:
        # sapwood depth in cm / 2cm for needle lenght =
        # portion of sensor in sapwood
        a = 0.5*swd
        # portion of sensor in inactive xylem
        b = 1.0-a
        # define weighted mean of T in the sapwood (a)
        # and T in the inactive xylem (b)
        Tsw  = (data_masked - (b*tmax)) / a

    # converts raw data [mV] into Sap Flux density [cm3*cm-2*h-1]
    SFD  = (0.0119 * ((tmax - Tsw)/Tsw)**1.231) * 3600.

    # if it was not masked then the original undef will be undef
    if isnotmask:
        SFD  = SFD.filled(undef)

    # 1D in = 1D out
    if isvec:
        SFD = np.squeeze(SFD)

    return SFD


if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)



#!/usr/bin/env python
from __future__ import division, absolute_import, print_function
import numpy as np


__all__ = ['jab']


def jab(arr, ind=None, nind=None, mask=None, weight=False, nsteps=1):
    """
    Jackknife-after-Bootstrap error, an approximation of the Jackknife error.

    If you one gives the indices of the bootstrap, a mask will be produced that
    tells if this sample was used in the current bootstrap:
        mask = np.zeros((nboot,nind), dtype=bool)
        for i in range(nboot):
            mask[i,ind[i,:]] = True
    One can also give this mask directly, reducing the memory amount
    by a factor of 64.


    Definition
    ----------
    def jab(arr, ind=None, nind=None, mask=None, weight=False, nsteps=1):


    Input
    -----
    arr       1D/2D-array of bootstrap standard error outputs.
              1st dim is number of bootstraps. Possible 2nd dim is number
              of bootstrap outputs.


    Optional Input
    --------------
    ind       2D-array of bootstrap indices. mask has priority over ind.
              1st dim is number of bootstraps. 2nd dim is number of samples
              per bootstrap.
    nind      # of data points, i.e. maximal jackknife index.
              Default: ind.shape[1], which assumes that each bootstraps took
              N samples from N data points.
              Ignored if mask is given.
    mask      2D-array of bollean if index used in this bootstrap.
              mask has priority over ind.
              1st dim is number of bootstraps. 2nd dim is number of samples
              per bootstrap.
    weight    if True: do weighted estimate after Wang et al. (1997)
              Default: False
    nsteps    Give nsteps estimates of JAB error, i.e. after each
              nboot/nsteps bootstraps.
              Default: 1


    Output
    ------
    Array of Jacknife-after-boostrap errors.
    This should converge towards the full Jackknif error after a large number
    of bootstraps.


    Restrictions
    ------------
    None


    References
    ----------
    Efron B. (1992)
        Jackknife-after-bootstrap standard errors and influence functions
        J. Royal. Statist. Soc. B 54, 83-127
    Wang et al. (1997)
        Weighted jackknife-after-bootstrap: A heuristic approach
        In Andradottir et al. (Eds.) Proceedings of the 29th conference on
        Winter simulation, 7-10 Dec 1997, Atlanta, Georgia, IEEE Computer
        Society Washington, DC, USA., pp. 240-245, doi:10.1145/268437.268486


    Example
    --------
    # 5 Bootstrap with 2 standard error outputs, e.g. for 2 parameters
    >>> import numpy as np
    >>> Bstat = np.array([[ 3.625,  5.375],
    ...                   [ 3.625,  4.125],
    ...                   [ 4.625,  3.75 ],
    ...                   [ 4.   ,  4.5  ],
    ...                   [ 4.25 ,  3.625]])

    # Indices of the bootstraps
    >>> indices = np.array([[0, 0, 4, 5, 5, 3, 2, 6],
    ...                     [1, 2, 3, 0, 5, 5, 4, 1],
    ...                     [3, 0, 3, 5, 4, 4, 1, 7],
    ...                     [0, 3, 5, 2, 2, 1, 4, 4],
    ...                     [4, 4, 5, 5, 3, 3, 2, 1]])

    # Normal JAB
    >>> from autostring import astr
    >>> print(astr(jab(Bstat, indices),3,pp=True))
    ['0.131' '0.393']

    # Normal JAB giving weights
    >>> mask = np.zeros((Bstat.shape[0],indices.shape[1]), dtype=bool)
    >>> for i in range(Bstat.shape[0]):
    ...     mask[i,indices[i,:]] = True
    >>> print(astr(jab(Bstat, mask=mask),3,pp=True))
    ['0.131' '0.393']

    # Weighted JAB
    >>> print(astr(jab(Bstat, indices, weight=True),3,pp=True))
    ['0.056' '0.168']

    # Normal JAB but bootstrap was done by taking 6 samples out of
    # 10 data points
    >>> print(astr(jab(Bstat, indices, nind=10),3,pp=True))
    ['0.146' '0.378']

    # Weighted JAB with 6 out of 10 bootstrap
    >>> print(astr(jab(Bstat[:,0], indices, nind=10),3,pp=True))
    0.146

    # Normal JAB, 2 steps
    >>> print(astr(jab(Bstat, indices, nsteps=5),3,pp=True))
    [['--   ' '0.000' '0.661' '0.312' '0.131']
     ['--   ' '0.000' '0.579' '0.288' '0.393']]


    License
    -------
    This file is part of the JAMS Python package, distributed under the MIT
    License. The JAMS Python package originates from the former UFZ Python
    library, Department of Computational Hydrosystems, Helmholtz Centre for
    Environmental Research - UFZ, Leipzig, Germany.

    Copyright (c) 2012-2021 Matthias Cuntz - mc (at) macu (dot) de

    Permission is hereby granted, free of charge, to any person obtaining a
    copy of this software and associated documentation files (the "Software"),
    to deal in the Software without restriction, including without limitation
    the rights to use, copy, modify, merge, publish, distribute, sublicense,
    and/or sell copies of the Software, and to permit persons to whom the
    Software is furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in
    all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.


    History
    -------
    Written,  Maren Goehler,  Aug 2012
    Modified, Matthias Cuntz, Nov 2012 - complete rewrite
              Matthias Cuntz, Dec 2012 - mask
              Matthias Cuntz, Feb 2013 - ported to Python 3
              Matthias Cuntz, Apr 2014 - assert
              Matthias Cuntz, Sep 2021 - code refactoring
    """
    # Shapes: make nboot bootstraps with nind samples per bootstrap
    # for nout parameters
    # arr[nboot, nout]
    # ind[nboot, nind]
    # mask[nboot, nind]

    # Check input
    onlyone = False
    if len(arr.shape) != 2:
        arr     = arr[:, np.newaxis]
        onlyone = True
    if mask is not None:
        assert len(mask.shape) == 2, 'Mask array must have 2 dimensions.'
        assert mask.shape[0] == arr.shape[0], 'First dimension of boostrap standard error array and mask array have to match.'
        nboot = mask.shape[0]
        nind  = mask.shape[1]
    else:
        assert len(ind.shape) == 2, 'Index array must have 2 dimensions.'
        assert ind.shape[0] == arr.shape[0], 'First dimension of boostrap standard error array and index array have to match.'
        nboot = ind.shape[0]
        if nind is not None:
            if nind < ind.shape[1]:
                print("Warning: nind < ind.shape[1]")
        else:
            nind = ind.shape[1]
    nout = arr.shape[1]

    if mask is None:
        # Produce mask with True where data point was used in bootstrap
        mask = np.zeros((nboot, nind), dtype=bool)
        for i in range(nboot):
            mask[i, ind[i, :]] = True

    seb_i      = np.ma.empty((nind, nout, nsteps))     # std jab_i
    seb_i.mask = np.zeros(seb_i.shape, dtype=bool)  # make mask an array
    b_i        = np.empty((nind, nsteps))              # number of indices
    step       = nboot // nsteps
    isnoti     = [np.empty((0), dtype=int) for i in range(nind)]
    for k in range(nsteps):
        # Searching the mask only for the current step and appending
        # the indices seems to be marginally faster than always searching
        # the mask from 0 to the current step
        #   imask   = mask[0:(k+1)*step,:]
        # and then later
        #   isnoti = np.where(imask[:,i]==False)[0]
        imask = mask[k*step:(k+1)*step, :]  # current step
        iarr  = arr[0:(k+1)*step, :]        # 1st to current step
        for i in range(nind):
            # Search all rows (in current step) where i-th column is False,
            # i.e. were not used in bootstrap
            iisnoti = k*step + np.where(imask[:, i] == False)[0]
            # Concat with all steps before
            isnoti[i] = np.concatenate((isnoti[i], iisnoti))
            # std and number of rows for possible weighting
            isize     = isnoti[i].size
            b_i[i, k] = isize
            if isize == 0:
                seb_i.mask[i, :, k] = True
            elif isize == 1:
                seb_i.mask[i, :, k] = True
            else:  # Wang et al. (1997), Eq. 3
                seb_i[i, :, k] = np.ma.std(iarr[isnoti[i], :], axis=0, ddof=0)

    # proposed empirical weight of Wang et al. (1997)
    if weight:  # Wang et al. (1997), Eq. 9
        w_i    = b_i.astype(float) / (
            float(np.sum(b_i, axis=0))/float(nind) + float(nind))
        seb_i *= w_i[:, np.newaxis, :]
    # jackknife-after-bootstrap error, Wang et al. (1997), Eq. 4
    se_jab = np.sqrt(float(nind-1)) * np.ma.std(seb_i, axis=0, ddof=0)

    if not np.any(se_jab.mask):
        se_jab = se_jab.data
    if onlyone:
        if nsteps == 1:
            return se_jab[0, 0]
        else:
            return se_jab[0, :]
    else:
        if nsteps == 1:
            return se_jab[:, 0]
        else:
            return se_jab


if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)

    # # 5 Bootstrap with 2 standard error outputs, e.g. for 2 parameters
    # Bstat = np.array([[ 3.625,  5.375],
    #                   [ 3.625,  4.125],
    #                   [ 4.625,  3.75 ],
    #                   [ 4.   ,  4.5  ],
    #                   [ 4.25 ,  3.625]])
    # # Indices of the bootstraps
    # indices = np.array([[0, 0, 4, 5, 5, 3, 2, 6],
    #                     [1, 2, 3, 0, 5, 5, 4, 1],
    #                     [3, 0, 3, 5, 4, 4, 1, 7],
    #                     [0, 3, 5, 2, 2, 1, 4, 4],
    #                     [4, 4, 5, 5, 3, 3, 2, 1]])

    # # Normal JAB
    # print jab(Bstat, indices)
    # #[ 0.13132114  0.39335039]

    # # Normal JAB giving weights
    # mask = np.zeros((Bstat.shape[0],indices.shape[1]), dtype=bool)
    # for i in range(Bstat.shape[0]):
    #     mask[i,indices[i,:]] = True
    # print jab(Bstat, mask=mask)
    # #[ 0.13132114  0.39335039]

    # # Weighted JAB
    # print jab(Bstat, indices, weight=True)
    # #[ 0.05603035  0.1678295 ]

    # # Normal JAB but bootstrap was done by taking 6 samples out of 10 data points
    # print jab(Bstat, indices, nind=10)
    # #[ 0.14620639  0.37764585]

    # # Weighted JAB with 6 out of 10 bootstrap
    # print jab(Bstat[:,0], indices, nind=10)
    # #0.146206392124

    # # Normal JAB, 5 steps
    # print jab(Bstat, indices, nsteps=5)
    # #[[-- 0.0 0.661437827766 0.311804782231 0.131321139382]
    # # [-- 0.0 0.578758099295 0.287799087529 0.393350393185]]



#!/usr/bin/env python
"""
Mimics Fortran intrinsic pack and unpack functions

This module was written by Matthias Cuntz while at Department of
Computational Hydrosystems, Helmholtz Centre for Environmental
Research - UFZ, Leipzig, Germany, and continued while at Institut
National de Recherche pour l'Agriculture, l'Alimentation et
l'Environnement (INRAE), Nancy, France.

:copyright: Copyright 2009-2023 Matthias Cuntz, see AUTHORS.rst for details.
:license: MIT License, see LICENSE for details.

.. moduleauthor:: Matthias Cuntz

The following functions are provided

.. autosummary::
   pack
   unpack

History
    * Written Jul 2009 by Matthias Cuntz (mc (at) macu (dot) de)
    * Ported to Python 3, Feb 2013, Matthias Cuntz
    * Assert matching mask and array dimensions, Apr 2014, Matthias Cuntz
    * Code refactoring, Sep 2021, Matthias Cuntz
    * Port to pyjams, May 2023, Matthias Cuntz
    * Rename value keyword to fill_value as in numpy masked arrays,
      May 2023, Matthias Cuntz
    * Assert output type equals input type, May 2023, Matthias Cuntz

"""
import numpy as np


__all__ = ['pack', 'unpack']


def pack(array, mask):
    """
    Mimics Fortran intrinsic pack (without optional vector)

    Packs the last dimensions of an arbitrary shaped array
    into a one dimensional array under a mask.

    The mask can have any number of dimensions up to the array dimension.

    Parameters
    ----------
    array : array
        ND-array to be packed
    mask : array
        Boolean ND-array with number of dimensions <= `array` dimensions

    Returns
    -------
    array
        `array` with reduced dimensions by number of mask dimensions minus one.
        Last dimension has only elements that correspond to elements of
        `mask == True`.

    Notes
    -----
    Result is undefined if all mask values are False.

    Examples
    --------

    Create some data for example an island in the middle of an ocean

    >>> import numpy as np
    >>> island = np.array([[0., 0., 0., 0., 0., 0., 0., 0., 0., 0.],
    ...                    [0., 0., 0., 1., 1., 1., 0., 0., 0., 0.],
    ...                    [0., 0., 0., 1., 1., 1., 0., 0., 0., 0.],
    ...                    [0., 0., 0., 1., 1., 1., 0., 0., 0., 0.],
    ...                    [0., 0., 0., 0., 0., 0., 0., 0., 0., 0.]])

    Pack array to keep only the elements of the island

    >>> mask = island == 1.0
    >>> pisland = pack(island, mask)
    >>> print(pisland.size, pisland.shape)
    9 (9,)
    >>> print(island.sum(), pisland.sum())
    9.0 9.0
    >>> print(pisland)
    [1. 1. 1. 1. 1. 1. 1. 1. 1.]

    Create data on the ocean and on the island for 2 time steps

    >>> tshape = (2, *island.shape)
    >>> temp = np.arange(np.prod(tshape)).reshape(tshape)

    Pack array to keep only the elements of the island

    >>> ptemp = pack(temp, mask)
    >>> print(ptemp.size, ptemp.shape)
    18 (2, 9)
    >>> print(ptemp)
    [[13 14 15 23 24 25 33 34 35]
     [63 64 65 73 74 75 83 84 85]]

    """
    dmask   = mask.shape
    ndmask  = np.ndim(mask)
    nmask   = mask.size
    darray  = array.shape
    ndarray = np.ndim(array)
    narray  = array.size

    # Check array and mask
    assert ndarray >= ndmask, (f'Input array has less dimensions {ndarray}'
                               f' then mask {ndmask}')
    k = 0
    while k > -ndmask:
        k -= 1
        assert dmask[k] == darray[k], (
            f'Input array and mask must have the same last dimensions.'
            f'Array: {darray}, Mask: {dmask}' )

    # Make array and mask 1d
    farray = array.ravel()  # flat in Fortran=column-major mode
    fmask  = mask.ravel()
    afmask = np.empty(narray, dtype=bool)
    nn = narray // nmask
    k  = 0
    while k < nn:
        afmask[k * nmask: (k + 1) * nmask] = fmask[:]
        k += 1

    # Mask array and reshape
    afarray = farray[afmask]
    dout = list(darray)
    k = 0
    while k < ndmask:
        del dout[-1]
        k += 1
    nnmask = int(mask.sum())
    dout.append(nnmask)
    out = np.reshape(afarray, dout)

    return out


def unpack(array, mask, fill_value=0):
    """
    Mimics Fortran intrinsic unpack (without optional field)

    Unpacks the last dimension into several dimensions under a mask.
    The unpacked elements will be set to a user-defined fill_value.
    The mask can have any number of dimensions up to the array dimensions.

    Parameters
    ----------
    array : array
        ND-array to be unpacked
    mask : array
        Boolean ND-array
    fill_value : float, optional
        Value of the new elements (default: 0)

    Returns
    -------
    array
        `array` having more dimensions by number of mask dimensions minus one.
        The new elements have the user-defined `fill_value`.

    Notes
    -----
    Result is undefined if all mask values are False.


    Examples
    --------

    Create some data for example an island in the middle of an ocean

    >>> import numpy as np
    >>> island = np.array([[0., 0., 0., 0., 0., 0., 0., 0., 0., 0.],
    ...                    [0., 0., 0., 1., 1., 1., 0., 0., 0., 0.],
    ...                    [0., 0., 0., 1., 1., 1., 0., 0., 0., 0.],
    ...                    [0., 0., 0., 1., 1., 1., 0., 0., 0., 0.],
    ...                    [0., 0., 0., 0., 0., 0., 0., 0., 0., 0.]])
    >>> print(island.size, island.shape)
    50 (5, 10)

    Pack array to keep only the elements of the island

    >>> mask = island == 1.0
    >>> pisland = pack(island, mask)
    >>> print(pisland.size, pisland.shape)
    9 (9,)

    Unpack the ocean and island data again

    >>> uisland = unpack(pisland, mask)
    >>> print(uisland.size, uisland.shape)
    50 (5, 10)
    >>> print(np.all(uisland == island))
    True

    Unpack the ocean and island data, filling ocean with -9

    >>> uisland = unpack(pisland, mask, -9)
    >>> print(uisland.size, uisland.shape)
    50 (5, 10)
    >>> print(uisland)
    [[-9. -9. -9. -9. -9. -9. -9. -9. -9. -9.]
     [-9. -9. -9.  1.  1.  1. -9. -9. -9. -9.]
     [-9. -9. -9.  1.  1.  1. -9. -9. -9. -9.]
     [-9. -9. -9.  1.  1.  1. -9. -9. -9. -9.]
     [-9. -9. -9. -9. -9. -9. -9. -9. -9. -9.]]

    Create data on the ocean and on the island for 2 time steps

    >>> tshape = (2, *island.shape)
    >>> temp = np.arange(np.prod(tshape)).reshape(tshape)

    >>> print(temp.size, temp.shape)
    100 (2, 5, 10)

    Pack array to keep only the elements of the island

    >>> ptemp = pack(temp, mask)
    >>> print(ptemp.size, ptemp.shape)
    18 (2, 9)

    Unpack the ocean and island data.
    Note the ocean will have the fill_value -1 not its original data.

    >>> utemp = unpack(ptemp, mask, fill_value=-1.)
    >>> print(utemp.size, utemp.shape)
    100 (2, 5, 10)
    >>> ii = np.where(utemp != -1.)
    >>> print(np.all(utemp[ii] == temp[ii]))
    True

    """
    dmask   = np.shape(mask)
    ndmask  = len(dmask)
    nmask   = mask.size
    darray  = np.shape(array)
    ndarray = len(darray)
    narray  = array.size

    # Check array and mask
    ntmask = int(mask.sum())
    assert darray[-1] == ntmask, (
        f'Last dimension of input array {darray[-1]} must have same size'
        f' as true values in mask {ntmask}.' )

    # Make multi mask array array
    icount = nmask
    for i in range(ndarray - 1):
        icount *= darray[i]
    mask1d = np.ravel(mask)
    masknd = np.empty(icount, dtype='bool')
    for i in range(icount // nmask):
        masknd[i * nmask:(i + 1) * nmask] = mask1d[:]

    # Make indeces
    index = np.arange(icount)
    ii = index[masknd]
    if len(ii) != narray:
        raise ValueError(f'Index creation failed. Index {len(ii)} Array'
                         f' {narray}.')

    # Flat output array
    array1d = np.ravel(array)
    arraynd = np.ones(icount, dtype=array.dtype) * fill_value
    arraynd[ii] = array1d[:]

    # Reshaped output array
    newdim = list(darray[0:-1])
    for i in range(ndmask):
        newdim.append(dmask[i])
    out = np.reshape(arraynd, newdim)

    return out


if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)



#!/usr/bin/env python
"""
Multi-dimensional non-parametric kernel regression

This module was written by Matthias Cuntz while at Department of
Computational Hydrosystems, Helmholtz Centre for Environmental
Research - UFZ, Leipzig, Germany, and continued while at Institut
National de Recherche pour l'Agriculture, l'Alimentation et
l'Environnement (INRAE), Nancy, France.

:copyright: Copyright 2012-2022 Matthias Cuntz, see AUTHORS.rst for details.
:license: MIT License, see LICENSE for details.

.. moduleauthor:: Matthias Cuntz

The following functions are provided

.. autosummary::
   kernel_regression_h
   kernel_regression

History
    * Written, Jun 2012 by  Matthias Cuntz - mc (at) macu (dot) de,
      inspired by Matlab routines of Yingying Dong, Boston College and
      Yi Cao, Cranfield University
    * Assert correct input, Apr 2014, Matthias Cuntz
    * Corrected bug in _boot_h: x.size->x.shape[0], Jan 2018, Matthias Cuntz
    * Code refactoring, Sep 2021, Matthias Cuntz
    * Use format strings, Apr 2022, Matthias Cuntz
    * Use minimize with method TNC instead of fmin_tnc,
      Apr 2022, Matthias Cuntz
    * Use helper function array2input to assure correct output type,
      Apr 2022, Matthias Cuntz
    * Return scalar h if 1-dimensional, Apr 2022, Matthias Cuntz
    * Output type is same as y instead of x or xout, Apr 2022, Matthias Cuntz

"""
import numpy as np
import scipy.optimize as opt
from .division import division
from .helper import array2input
# from division import division
# from helper import array2input


__all__ = ['kernel_regression_h', 'kernel_regression']


def _nadaraya_watson(z, y):
    """
    Helper function that calculates the Nadaraya-Watson estimator
    for a given kernel.
    Until now there is only the gaussian kernel.

    """
    kerf = 1. / np.sqrt(2. * np.pi) * np.exp(-0.5 * z * z)
    w = np.prod(kerf, 1)
    out = division(np.dot(w, y), np.sum(w), np.nan)

    return out


def _cross_valid_h(h, x, y):
    """
    Helper function that calculates cross-validation function for the
    Nadaraya-Watson estimator, which is basically the mean square error
    where model estimate is replaced by the jackknife estimate
    (Hardle and Muller 2000).

    """
    n = x.shape[0]
    # allocate output
    out = np.empty(n)
    # Loop through each regression point
    for i in range(n):
        # all-1 points
        xx = np.delete(x, i, axis=0)
        yy = np.delete(y, i, axis=0)
        z = (xx - x[i, :]) / h
        out[i] = _nadaraya_watson(z, yy)
    cv = np.sum((y - out)**2) / float(n)

    return cv


def _boot_h(h, x, y):
    """
    Helper function that calculates bootstrap function for the
    Nadaraya-Watson estimator, which is basically the mean square error
    where model estimate is replaced by the jackknife estimate
    (Hardle and Muller 2000).

    This does basically _cross_valid_h for 100 random points.

    """
    n = 100
    ind = np.random.randint(x.shape[0], size=n)
    # allocate output
    out = np.empty(n)
    # Loop through each bootstrap point
    for i in range(n):
        # all-1 points
        xx = np.delete(x, i, axis=0)
        yy = np.delete(y, i, axis=0)
        z = (xx - x[i, :]) / h
        out[i] = _nadaraya_watson(z, yy)
    cv = np.sum((y[ind] - out)**2) / float(n)

    return cv


def kernel_regression_h(x, y, silverman=False):
    """
    Optimal bandwidth for multi-dimensional non-parametric kernel regression

    Optimal bandwidth is determined using cross-validation or
    Silverman's rule-of-thumb.

    Parameters
    ----------
    x : array_like (n, k)
        Independent values
    y : array_like (n)
        Dependent values
    silverman : bool, optional
        Use Silverman's rule-of-thumb to calculate bandwidth *h* if True,
        otherwise determine *h* via cross-validation

    Returns
    -------
    float or array
        Optimal bandwidth *h*. If multidimensional regression then *h* is a
        1d-array, assuming a diagonal bandwidth matrix.

    References
    ----------
    Hardle W and Muller M (2000) Multivariate and semiparametric kernel
       regression. In MG Schimek (Ed.), Smoothing and regression: Approaches,
       computation, and application (pp. 357-392). Hoboken, NJ, USA: John Wiley
       & Sons, Inc. doi: 10.1002/9781118150658.ch12


    Examples
    --------
    >>> n = 10
    >>> x = np.zeros((n, 2))
    >>> x[:, 0] = np.arange(n, dtype=float) / float(n-1)
    >>> x[:, 1] = 1. / (np.arange(n, dtype=float) / float(n-1) + 0.1)
    >>> y = 1. + x[:, 0]**2 - np.sin(x[:, 1])**2

    >>> h = kernel_regression_h(x, y)
    >>> print(np.allclose(h, [0.172680, 9.516907], atol=0.0001))
    True

    >>> h = kernel_regression_h(x, y, silverman=True)
    >>> print(np.allclose(h, [0.229190, 1.903381], atol=0.0001))
    True

    >>> n = 10
    >>> x = np.arange(n, dtype=float) / float(n-1)
    >>> y = 1. + x**2 - np.sin(x)**2

    >>> h = kernel_regression_h(x, y)
    >>> print(np.around(h, 4))
    0.045

    >>> h = kernel_regression_h(x, y, silverman=True)
    >>> print(np.around(h, 4))
    0.2248

    """
    xx = np.array(x)
    yy = np.array(y)
    # Check input
    assert xx.shape[0] == yy.size, (
        f'size(x, 0) != size(y): {xx.shape[0]} != {yy.size}' )
    if xx.ndim == 1:  # deal with 1d-arrays
        xx = xx[:, np.newaxis]
    n = xx.shape[0]
    d = xx.shape[1]

    # Silverman (1986), Scott (1992), Bowman and Azzalini (1997)
    # Very similar to stats.gaussian_kde
    # h has dimension d
    h = ( (4. / float(d + 2) / float(n))**(1. / float(d + 4)) *
          np.std(xx, axis=0, ddof=1) )

    if not silverman:
        # Find the optimal h
        bounds = [(0.2*i, 5.0*i) for i in h]
        if n <= 100:
            res = opt.minimize(
                _cross_valid_h, h, args=(xx, yy), method='TNC', bounds=bounds,
                options={'ftol': 1e-10, 'xtol': 1e-10, 'maxfun': 1000})
            h = res.x
        else:
            res = opt.minimize(
                _boot_h, h, args=(xx, yy), method='TNC', bounds=bounds,
                options={'ftol': 1e-10, 'xtol': 1e-10, 'maxfun': 1000})
            h = res.x

    if len(h) == 1:
        h = h[0]

    return h


def kernel_regression(x, y, h=None, silverman=False, xout=None):
    """
    Multi-dimensional non-parametric kernel regression

    Optimal bandwidth can be estimated by cross-validation
    or by using Silverman's rule-of-thumb.

    Parameters
    ----------
    x : array_like (n, k)
        Independent values
    y : array_like (n)
        Dependent values
    h : float or array_like(k), optional
        Use *h* as bandwidth for calculating regression values if given,
        otherwise determine optimal *h* using cross-validation or by using
        Silverman's rule-of-thumb if *silverman==True*.
    silverman : bool, optional
        Use Silverman's rule-of-thumb to calculate bandwidth *h* if True,
        otherwise determine *h* via cross-validation.
        Only used if *h* is not given.
    xout : ndarray(n, k), optional
        Return fitted values at *xout* if given,
        otherwise return fitted values at *x*.

    Returns
    -------
    array_like with same type as *x*, or *xout* if given
        Fitted values at *x*, or at *xout* if given

    References
    ----------
    Hardle W and Muller M (2000) Multivariate and semiparametric kernel
       regression. In MG Schimek (Ed.), Smoothing and regression: Approaches,
       computation, and application (pp. 357-392). Hoboken, NJ, USA: John Wiley
       & Sons, Inc. doi: 10.1002/9781118150658.ch12

    Examples
    --------
    >>> n = 10
    >>> x = np.zeros((n, 2))
    >>> x[:, 0] = np.arange(n, dtype=float) / float(n-1)
    >>> x[:, 1] = 1. / (np.arange(n, dtype=float) / float(n-1) + 0.1)
    >>> y = 1. + x[:, 0]**2 - np.sin(x[:, 1])**2

    Separate determination of h and kernel regression

    >>> h = kernel_regression_h(x, y)
    >>> yk = kernel_regression(x, y, h)
    >>> print(np.allclose(yk[0:6],
    ...       [0.52241, 0.52570, 0.54180, 0.51781, 0.47644, 0.49230],
    ...       atol=0.0001))
    True

    Single call to kernel regression

    >>> yk = kernel_regression(x, y)
    >>> print(np.allclose(yk[0:6],
    ...       [0.52241, 0.52570, 0.54180, 0.51781, 0.47644, 0.49230],
    ...       atol=0.0001))
    True

    Single call to kernel regression using Silverman's rule-of-thumb for h

    >>> yk = kernel_regression(x, y, silverman=True)
    >>> print(np.allclose(yk[0:6],
    ...       [0.691153, 0.422809, 0.545844, 0.534315, 0.521494, 0.555426],
    ...       atol=0.0001))
    True

    >>> n = 5
    >>> xx = np.empty((n, 2))
    >>> xx[:, 0] = (np.amin(x[:, 0]) + (np.amax(x[:, 0]) - np.amin(x[:, 0])) *
    ...                                 np.arange(n, dtype=float) / float(n))
    >>> xx[:, 1] = (np.amin(x[:, 1]) + (np.amax(x[:, 1]) - np.amin(x[:, 1])) *
    ...                                 np.arange(n, dtype=float) / float(n))
    >>> yk = kernel_regression(x, y, silverman=True, xout=xx)
    >>> print(np.allclose(yk,
    ...       [0.605485, 0.555235, 0.509529, 0.491191, 0.553325],
    ...       atol=0.0001))
    True

    """
    xx = np.array(x)
    yy = np.array(y)
    # Check input
    assert xx.shape[0] == yy.size, (
        f'size(x,0) != size(y): {xx.shape[0]} != {yy.size}' )
    # deal with 1d-arrays and save 1d input type
    if xx.ndim == 1:
        xx = xx[:, np.newaxis]
    d = xx.shape[1]

    # determine h
    if h is None:
        hh = kernel_regression_h(xx, yy, silverman=silverman)
    else:
        if np.size(np.shape(h)) == 0:
            hh = np.repeat(h, d)  # use h for all dimensions if one h given
        else:
            hh = np.array(h)
        assert np.size(hh) == d, 'size(h) must be 1 or x.shape[1]'

    # Calc regression
    if xout is None:
        xxout = xx
    else:
        xxout = np.array(xout)
    if xxout.ndim == 1:
        xxout = xxout[:, np.newaxis]
    nout  = xxout.shape[0]
    dout  = xxout.shape[1]
    assert d == dout, f'size(x, 1) != size(xout, 1): {d} != {dout}'
    # allocate output
    out = np.empty(nout)
    # Loop through each regression point
    for i in range(nout):
        # scaled deference from regression point
        z = (xx - xxout[i, :]) / hh
        # nadaraya-watson estimator of gaussian multivariate kernel
        out[i] = _nadaraya_watson(z, y)

    out = array2input(out, y, undef=np.nan)

    return out


if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)

    # nn = 1000
    # x = np.zeros((nn, 2))
    # x[:, 0] = np.arange(nn, dtype=float) / float(nn-1)
    # x[:, 1] = 1. / (x[:, 0] + 0.1)
    # y      = 1. + x[:, 0]**2 - np.sin(x[:, 1])**2
    # h = kernel_regression_h(x, y)
    # print(h)
    # yy = kernel_regression(x, y, h, xout=x)
    # print(yy[0], yy[-1])
    # print(kernel_regression(x, y))
    # h = kernel_regression_h(x, y, silverman=True)
    # print(h)
    # print(kernel_regression(x, y, h))
    # ss = np.shape(x)
    # nn = 5
    # xx = np.empty((nn, ss[1]))
    # xx[:,0] = (np.amin(x[:,0]) + (np.amax(x[:,0])-np.amin(x[:,0])) *
    #                               np.arange(nn,dtype=float)/float(nn))
    # xx[:,1] = (np.amin(x[:,1]) + (np.amax(x[:,1])-np.amin(x[:,1])) *
    #                               np.arange(nn,dtype=float)/float(nn))
    # print(kernel_regression(x, y, h, xout=xx))



#!/usr/bin/env python
import numpy as np
from pyjams.const import eps


__all__ = ['around']


def around(num, powten, ceil=False, floor=False):
    """
        Round to the passed power of ten.


        Definition
        ----------
        def around(num, powten=None, ceil=False, floor=False):


        Input
        -----
        num        ND array of floats


        Optional Input
        --------------
        powten     Power of ten array
                   If missing, simple round (ceil, floor) is taken.
        ceil       ceil instead of round to the nearest power of ten
        floor      floor instead of round to the nearest power of ten


        Output
        ------
        Rounded values.


        Restrictions
        ------------
        Powten is exactly opposite of decimal keyword of numpy.around.

        From numpy.around documentation:
        'For values exactly halfway between rounded decimal values,
        Numpy rounds to the nearest even value. Thus 1.5 and 2.5 round to 2.0,
        -0.5 and 0.5 round to 0.0, etc. Results may also be surprising due to the
        inexact representation of decimal fractions in the IEEE floating point
        standard and errors introduced when scaling by powers of ten.'


        Examples
        --------
        >>> from jams import astr
        >>> print(astr(around(np.array([3.5967,345.5967]), -3),3,pp=True))
        ['  3.597' '345.597']

        >>> print(astr(around(np.array([1994344,345.5967]), [3,-3]),3,pp=True))
        ['1.994e+06' '3.456e+02']

        >>> print(astr(around(np.array([1994344,345.5967]), [3,-3], ceil=True),3,pp=True))
        ['1.995e+06' '3.456e+02']

        >>> print(astr(around(np.array([1994344,345.5967]), [3,-3], floor=True),3,pp=True))
        ['1.994e+06' '3.456e+02']

        >>> print(astr(around(np.array([3.5967,345.5967]), 3),3,pp=True))
        ['0.000' '0.000']

        >>> print(astr(around(np.array([3.5967,345.5967]), 3, ceil=True),3,pp=True))
        ['1000.000' '1000.000']


        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT
        License. The JAMS Python package originates from the former UFZ Python library,
        Department of Computational Hydrosystems, Helmholtz Centre for Environmental
        Research - UFZ, Leipzig, Germany.

        Copyright (c) 2011-2016 Matthias Cuntz - mc (at) macu (dot) de

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  MC, Jun 2011
        Modified, MC, Feb 2013 - ported to Python 3
                  MC, Apr 2014 - assert
                  MC, Nov 2016 - const.tiny -> const.eps
    """
    #
    # Check input
    assert (ceil+floor) < 2, 'ceil and floor keywords given.'
    if (powten == None):
        ipowten = 0
    else:
        ipowten = np.array(powten)
    nnum = np.size(num)
    npowten = np.size(ipowten)
    assert not ((npowten != nnum) and (npowten != 1)), 'powten must be scalar or have array size of input numbers.'
    #
    # Shift decimal point
    # Does not work, too imprecise: out = num * np.exp(-ipowten*np.log(10.))
    out = num * 10.**(-ipowten)
    # Round/ceil/floor
    if (ceil):
        out = np.ceil(out-10.*eps)
    elif (floor):
        out = np.floor(out+10.*eps)
    else:
       out = np.around(out)
    # Shift back decimal point
    # Does not work, too imprecise: out *= np.exp(ipowten*np.log(10.))
    out *= 10.**ipowten

    return out


if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)



#!/usr/bin/env python
import numpy as np
from scipy import interpolate, signal


def xkcd_line(x, y, xlim=None, ylim=None, mag=1.0, f1=30, f2=0.05, f3=15):
    """
        Mimic a hand-drawn line from (x, y) data


        Definition
        ----------
        def xkcd_line(x, y, xlim=None, ylim=None, mag=1.0, f1=30, f2=0.05, f3=15):


        Input
        -----
        x, y          array_like; arrays to be modified


        Optional Input
        --------------
        xlim, ylim    data range; the assumed plot range for the modification.
                      If not specified,  they will be guessed from the  data
        mag           float; the magnitude of the distortion (default: 1.0)
        f1, f2, f3    int, float, int; filtering parameters.
                      f1 gives the size of the window (default: 50)
                      f2 gives the high-frequency cutoff (default: 0.01)
                      f3 gives the size of the filter (default: 15)


        Output
        ------
        x, y          ndarrays; the modified lines


        References
        ----------
        See xkcd below.


        Examples
        --------
        for line in ax.lines:
            x, y         = line.get_data()
            x_int, y_int = xkcd_line(x, y, xlim, ylim, mag, f1, f2, f3)
            line.set_data(x_int, y_int)


        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT
        License. The JAMS Python package originates from the former UFZ Python library,
        Department of Computational Hydrosystems, Helmholtz Centre for Environmental
        Research - UFZ, Leipzig, Germany.

        Copyright (c) 2013-2019 Matthias Cuntz

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.

        History
        -------
        Written,  MC, Mar 2013
    """
    # assure array
    x = np.asarray(x)
    y = np.asarray(y)

    # get limits for rescaling
    if xlim is None: xlim = (x.min(), x.max())
    if ylim is None: ylim = (y.min(), y.max())
    if xlim[1] == xlim[0]: xlim = ylim
    if ylim[1] == ylim[0]: ylim = xlim

    # scale the data
    x_scaled = (x - xlim[0]) * 1. / (xlim[1] - xlim[0])
    y_scaled = (y - ylim[0]) * 1. / (ylim[1] - ylim[0])

    # compute the total distance along the path
    dx       = x_scaled[1:] - x_scaled[:-1]
    dy       = y_scaled[1:] - y_scaled[:-1]
    dist_tot = np.sum(np.sqrt(dx*dx + dy*dy))

    # number of interpolated points is proportional to the distance
    Nu = int(200 * dist_tot)
    u  = np.arange(-1, Nu + 1) * 1. / (Nu - 1)

    # interpolate curve at sampled points
    # k            = min(3, len(x) - 1)
    k            = min(3, x.size - 1)
    res          = interpolate.splprep([x_scaled, y_scaled], s=0, k=k)
    x_int, y_int = interpolate.splev(u, res[0])

    # we perturb perpendicular to the drawn line
    dx   = x_int[2:] - x_int[:-2]
    dy   = y_int[2:] - y_int[:-2]
    # horizontal or vertical lines
    # np.sign(np.cumsum(np.random.random(dx.size)-0.5)) emulates something like a Brownian motion
    # i.e. auto-correlated random walks around 0; just the sign interests here.
    eps = np.maximum(np.abs(np.amax(x_scaled)-np.amin(x_scaled)), np.abs(np.amax(y_scaled)-np.amin(y_scaled)))/Nu
    if np.all(np.abs(dx) < eps):
        dx = np.sign(np.cumsum(np.random.random(dx.size)-0.5)) * eps
    if np.all(np.abs(dy) < eps):
        dy = np.sign(np.cumsum(np.random.random(dx.size)-0.5)) * eps
    # equal distances
    if np.all(np.sign(dx) == np.sign(dx[0])):
        dx *= np.sign(np.cumsum(np.random.random(dx.size)-0.5))
    if np.all(np.sign(dy) == np.sign(dy[0])):
        dy *= np.sign(np.cumsum(np.random.random(dx.size)-0.5))
    dist = np.sqrt(dx * dx + dy * dy)

    # create a filtered perturbation
    # coeffs       = mag * np.random.normal(0, 0.01, len(x_int) - 2)
    coeffs       = mag * np.random.normal(0, 0.01, x_int.size - 2)
    b            = signal.firwin(f1, f2*dist_tot, window=('kaiser', f3))
    response     = signal.lfilter(b, 1, coeffs)
    x_int[1:-1] += response * dy / dist
    y_int[1:-1] += response * dx / dist

    # un-scale data
    x_int = x_int[1:-1] * (xlim[1] - xlim[0]) + xlim[0]
    y_int = y_int[1:-1] * (ylim[1] - ylim[0]) + ylim[0]

    return x_int, y_int


def xkcd(ax,
         mag=1.0,
         f1=50, f2=0.01, f3=15,
         bgcolor='w',
         title_size=None,
         xaxis_loc=None, yaxis_loc=None,
         xaxis_arrow='+', yaxis_arrow='+',
         ax_extend=0.1,
         xlabel_inside=0., ylabel_inside=0.,
         ticks=False,
         xticks_inside=0., yticks_inside=0.,
         ):
    """
        Make axis look hand-drawn

        This adjusts all lines, text, legends, and axes in the figure to look
        like xkcd plots, a webcomic from Randall Munroe. Other plot elements are not modified.


        Definition
        ----------
        def xkcd(ax,
                 mag=1.0,
                 f1=50, f2=0.01, f3=15,
                 bgcolor='w',
                 title_size=None,
                 xaxis_loc=None, yaxis_loc=None,
                 xaxis_arrow='+', yaxis_arrow='+',
                 ax_extend=0.1,
                 xlabel_inside=0., ylabel_inside=0.,
                 ticks=False,
                 xticks_inside=0., yticks_inside=0.,
                 ):


        Input
        -----
        ax    Axes instance the axes instance to be modified.


        Optional Input
        --------------

        mag                            float; the magnitude of the distortion (default: 1.0)
        f1, f2, f3                     int, float, int; filtering parameters.
                                       f1 gives the size of the window (default: 50)
                                       f2 gives the high-frequency cutoff (default: 0.01)
                                       f3 gives the size of the filter (default: 15)
        bgcolor                        str; color around lines so that axis look brocken,
                                       i.e. lines are overdrawn on axis (default: 'w')
        titel_size                     float; poitn size of plot title. If None, same size as axis labels.
                                       (default: None)
        xaxis_loc, yaxis_log           float; The locations to draw the x and y axes in data coordinates.
                                       If not specified, they will be drawn from the bottom left of the plot.
                                       (default: None)
        xaxis_arrow, yaxis_arrow       str; where to draw arrows on the x/y axes
                                       Options are '+', '-', '+-', or '' (default: '+')
        ax_extend                      float; How far (fractionally) to extend the drawn axes beyond
                                       the original axes limits (default: 0.1)
        xlabel_inside, ylabel_inside   float: By how much the labels are shifted (default: 0.0)

        The last two options are not working how with mc_plot_template
        ticks                          True: change tick labels; False: no tick labels are drawn (default: False)
        xticks_inside, yticks_inside   float: By how much the ticks are shifted (default: 0.0)


        Output
        ------
        ax is basically empty and all former elements are redrawn on plot.


        Note
        ----
        For reproducible plots, seed the random number generator before each new plot.
        If a new line was added, the old lines will look the same. The legend will be different though.


        References
        ----------
        This is the modified XKCD plot generator of Jake Vanderplas
        http://nbviewer.ipython.org/url/jakevdp.github.com/downloads/notebooks/XKCD_plots.ipynb

        The idea for this comes from work by Damon McDougall
        http://www.mail-archive.com/matplotlib-users@lists.sourceforge.net/msg25499.html


        Examples
        --------
        import matplotlib.pylab as plt
        fig = plt.figure(1)
        ax = fig.add_axes([0.1,0.1,0.5,0.5])
        ax.plot(range(10), label='Line')
        ax.set_title('Title')
        ax.set_xlabel('x label')
        ax.set_ylabel('y label')
        ax.legend()
        xkcd(ax)


        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT License.

        Copyright (c) 2013 Matthias Cuntz - mc (at) macu (dot) de

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  MC, Mar 2013
    """
    import matplotlib.pylab as plt
    import matplotlib.font_manager as fm

    # remember random state for later resetting
    random_state = np.random.get_state()

    # Get axes aspect
    ext = ax.get_window_extent().extents
    aspect = (ext[3] - ext[1]) / (ext[2] - ext[0])

    xlim = ax.get_xlim()
    ylim = ax.get_ylim()

    xspan = xlim[1] - xlim[0]
    yspan = ylim[1] - xlim[0]

    xax_lim = (xlim[0] - ax_extend * xspan,
               xlim[1] + ax_extend * xspan)
    yax_lim = (ylim[0] - ax_extend * yspan,
               ylim[1] + ax_extend * yspan)

    if xaxis_loc is None: xaxis_loc = ylim[0]
    if yaxis_loc is None: yaxis_loc = xlim[0]

    # Draw axes
    acolor  = ax.get_xaxis().get_gridlines()[0].get_color()
    xaxis = plt.Line2D([xax_lim[0], xax_lim[1]], [xaxis_loc, xaxis_loc],
                      linestyle='-', color=acolor)
    yaxis = plt.Line2D([yaxis_loc, yaxis_loc], [yax_lim[0], yax_lim[1]],
                      linestyle='-', color=acolor)

    # adjust the axes
    if ticks:
        for x, xtext in zip(ax.get_xticks(), ax.get_xticklabels()):
            ax.text(x, xaxis_loc - 0.08 * yspan * (2 * xticks_inside - 1), xtext.get_text(),
                fontsize=xtext.get_size(), ha='center', va='bottom' if xticks_inside else 'top', rotation=0)
        for y, ytext in zip(ax.get_yticks(), ax.get_yticklabels()):
            ax.text(yaxis_loc + 0.02 * xspan * (2 * yticks_inside - 1), y, ytext.get_text(),
                fontsize=ytext.get_size(), ha='left' if yticks_inside else 'right', va='center', rotation=0)

    # Label axes
    siz = ax.get_xaxis().get_label().get_size()
    ax.text(xax_lim[1], xaxis_loc - 0.2 * yspan * (2 * xlabel_inside - 1), ax.get_xlabel(),
            fontsize=siz, ha='right', va='bottom' if xlabel_inside else 'top', rotation=0)
    ax.text(yaxis_loc + 0.04 * xspan * (2 * ylabel_inside - 1), yax_lim[1], ax.get_ylabel(),
            fontsize=siz, ha='right', va='bottom' if ylabel_inside else 'top', rotation=84)

    # Title - default: same size as axis labels
    if title_size is not None:
        siz2 = title_size
    else:
        siz2 = siz
    ax.text(0.5 * (xax_lim[1] + xax_lim[0]), yax_lim[1], ax.get_title(),
            ha='center', va='bottom', fontsize=siz2)

    # Draw arrow-heads at the end of axes lines
    arr1       = 0.04 * np.array([-1, 0, -1])
    arr2       = 0.03 * np.array([-1, 0, 1])
    arr1[::2] += np.random.normal(0, 0.005 / 2, 2)
    arr2[::2] += np.random.normal(0, 0.005 / 2, 2)
    x, y = xaxis.get_data()
    if '+' in str(xaxis_arrow):
        ax.plot(x[-1] + arr1 * xspan * aspect,
                y[-1] + arr2 * yspan,
                color=acolor, lw=2)
    if '-' in str(xaxis_arrow):
        ax.plot(x[0] - arr1 * xspan * aspect,
                y[0] - arr2 * yspan,
                color=acolor, lw=2)

    x, y = yaxis.get_data()
    if '+' in str(yaxis_arrow):
        ax.plot(x[-1] + arr2 * xspan * aspect**2,
                y[-1] + arr1 * yspan / aspect,
                color=acolor, lw=2)
    if '-' in str(yaxis_arrow):
        ax.plot(x[0] - arr2 * xspan * aspect**2,
                y[0] - arr1 * yspan / aspect,
                color=acolor, lw=2)

    # Set the axis limits
    ax.set_xlim(xax_lim[0] - 0.1 * xspan,
                xax_lim[1] + 0.1 * xspan)
    ax.set_ylim(yax_lim[0] - 0.1 * yspan,
                yax_lim[1] + 0.1 * yspan)

    # The lines
    Nlines = len(ax.lines)
    lines  = [xaxis, yaxis] + [ax.lines.pop(0) for i in range(Nlines)]
    for line in lines:
        x, y = line.get_data()
        ls   = line.get_linestyle()
        if ls != 'None':
            x_int, y_int = xkcd_line(x, y, xlim, ylim, mag, f1, f2, f3)
        else:
            x_int, y_int = x, y
        # create foreground and background line
        lw = line.get_linewidth()
        line.set_linewidth(2*lw)
        line.set_data(x_int, y_int)

        # White surrounding of line makes them look overplot on axis
        if (line is not xaxis) and (line is not yaxis) and ls != 'None':
            line_bg = plt.Line2D(x_int, y_int, color=bgcolor, linewidth=2*lw+4)
            ax.add_line(line_bg)
        ax.add_line(line)

    # Change all the fonts to humor-sans.
    import os
    fhumor = os.path.join(os.path.dirname(__file__), 'Humor-Sans.ttf') # in jams_python/jams
    for text in ax.texts:
        tsize  = text.get_size()
        prop = fm.FontProperties(fname=fhumor, size=tsize)
        text.set_fontproperties(prop)

    # modify legend
    leg = ax.get_legend()
    if leg is not None:
        np.random.set_state(random_state) # restate random number generator for reproducible results
        leg.set_frame_on(False)
        for child in leg.get_children():
            if isinstance(child, plt.Line2D):
                x, y = child.get_data()
                child.set_data(xkcd_line(x, y, mag=10.*mag, f1=2*f1, f2=f2/10.))
                child.set_linewidth(2*child.get_linewidth())
            if isinstance(child, plt.Text):
                tsize  = child.get_size()
                prop = fm.FontProperties(fname=fhumor, size=tsize)
                child.set_fontproperties(prop)

    # remove standard axis
    ax.set_title('')
    ax.set_xlabel('')
    ax.set_ylabel('')
    ax.set_xticks([])
    ax.set_yticks([])
    ax.set_axis_off()

    return ax


if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)



#!/usr/bin/env python
"""
This is the unittest for division module.

python -m unittest -v tests/test_division.py
python -m pytest --cov=pyjams --cov-report term-missing -v tests/test_division.py

"""
import unittest


class TestDivision(unittest.TestCase):
    """
    Tests for division.py
    """

    def test_division(self):
        import numpy as np
        import pandas as pd
        from pyjams import division, div

        for dfunc in [division, div]:
            a = (1., 2., 3.)
            b = 2.
            assert isinstance(dfunc(a, b), tuple)
            self.assertEqual(list(dfunc(a, b)),
                             [0.5, 1.0, 1.5])

            a = [1., 2., 3.]
            b = 2.
            assert isinstance(dfunc(a, b), list)
            self.assertEqual(list(dfunc(a, b)),
                             [0.5, 1.0, 1.5])

            a = np.array([1., 2., 3.])
            b = 2.
            assert isinstance(dfunc(a, b), np.ndarray)
            self.assertEqual(list(dfunc(a, b)),
                             [0.5, 1.0, 1.5])

            # pandas.Series
            a = np.array([1., 2., 3.])
            b = 2.
            d1 = [pd.to_datetime('2020-06-01 12:30'),
                  pd.to_datetime('2020-09-03 16:00'),
                  pd.to_datetime('2021-06-03 11:00')]
            df = pd.Series(a)
            df.index = d1
            dd = dfunc(df, b)
            assert isinstance(dd, pd.Series)
            self.assertEqual(list(dd.values), [0.5, 1.0, 1.5])

            # pandas.DataFrame
            a = np.array([1., 2., 3.])
            a1 = np.vstack([a, a])
            d1 = [pd.to_datetime('2020-06-01 12:30'),
                  pd.to_datetime('2020-09-03 16:00')]
            df = pd.DataFrame(a1)
            df.index = d1
            dd = dfunc(df, b)
            assert isinstance(dd, pd.DataFrame)
            self.assertEqual(list(dd.values.flatten()),
                             [0.5, 1.0, 1.5, 0.5, 1.0, 1.5])

            a = 1.
            b = (1., 2., 4.)
            assert isinstance(dfunc(a, b), tuple)
            self.assertEqual(list(dfunc(a, b)),
                             [1.0, 0.5, 0.25])

            a = 1.
            b = [1., 2., 4.]
            assert isinstance(dfunc(a, b), list)
            self.assertEqual(list(dfunc(a, b)),
                             [1.0, 0.5, 0.25])

            a = 1.
            b = np.array([1., 2., 4.])
            assert isinstance(dfunc(a, b), np.ndarray)
            self.assertEqual(list(dfunc(a, b)),
                             [1.0, 0.5, 0.25])

            a = 1.
            b = np.array([1., 2., 4.])
            d1 = [pd.to_datetime('2020-06-01 12:30'),
                  pd.to_datetime('2020-09-03 16:00'),
                  pd.to_datetime('2021-06-03 11:00')]
            df = pd.Series(b)
            df.index = d1
            dd = dfunc(a, df)
            assert isinstance(dd, pd.Series)
            self.assertEqual(list(dd.values), [1.0, 0.5, 0.25])

            a = 1.
            b = np.array([1., 2., 4.])
            b1 = np.vstack([b, b])
            d1 = [pd.to_datetime('2020-06-01 12:30'),
                  pd.to_datetime('2020-09-03 16:00')]
            df = pd.DataFrame(b1)
            df.index = d1
            dd = dfunc(a, df)
            assert isinstance(dd, pd.DataFrame)
            self.assertEqual(list(dd.values.flatten()),
                             [1.0, 0.5, 0.25, 1.0, 0.5, 0.25])

            a = [1., 1., 1.]
            b = [2., 1., 0.]
            dd = dfunc(a, b)
            assert isinstance(dd, list)
            self.assertEqual(list(dd[0:2]), [0.5, 1.0])
            assert np.isnan(dd[2])
            self.assertEqual(list(dfunc(a, b, 0.)),
                             [0.5, 1.0, 0.0])
            self.assertEqual(list(dfunc(a, b, otherwise=0.)),
                             [0.5, 1.0, 0.0])
            dd = dfunc(a, b, prec=1.)
            self.assertEqual(dd[0], 0.5)
            assert np.isnan(dd[1])
            assert np.isnan(dd[2])

            a = np.array([1., 1., 1.])
            b = [2., 1., 0.]
            dd = dfunc(a, b)
            assert isinstance(dd, np.ndarray)
            self.assertEqual(list(dd[0:2]), [0.5, 1.0])
            assert np.isnan(dd[2])

            a = [1., 1., 1.]
            b = [2., 1., 0.]
            d1 = [pd.to_datetime('2020-06-01 12:30'),
                  pd.to_datetime('2020-09-03 16:00'),
                  pd.to_datetime('2021-06-03 11:00')]
            df = pd.Series(a)
            df.index = d1
            dd = dfunc(df, b)
            assert isinstance(dd, pd.Series)
            self.assertEqual(list(dd.values[:2]), [0.5, 1.0])
            assert np.isnan(dd.iloc[2])

            a = [1., 1., 1.]
            b = [2., 1., 0.]
            d1 = [pd.to_datetime('2020-06-01 12:30'),
                  pd.to_datetime('2020-09-03 16:00'),
                  pd.to_datetime('2021-06-03 11:00')]
            df = pd.Series(b)
            df.index = d1
            dd = dfunc(a, df)
            assert isinstance(dd, list)
            self.assertEqual(dd[:2], [0.5, 1.0])
            assert np.isnan(dd[2])

            a = np.array([1., 1., 1.])
            b = np.array([2., 1., 0.])
            dd = dfunc(a, b)
            assert isinstance(dd, np.ndarray)
            self.assertEqual(list(dd[0:2]), [0.5, 1.0])
            assert np.isnan(dd[2])

            mask = [0, 0, 1]
            a = np.ma.array([1., 1., 1.], mask=mask)
            b = np.array([2., 1., 0.])
            dd = dfunc(a, b)
            assert isinstance(dd, np.ma.masked_array)
            self.assertEqual(list(dd), [0.5, 1.0, np.ma.masked])

            a = np.array([1., 1., 1.])
            mask = [0, 0, 1]
            b = np.ma.array([2., 1., 0.], mask=mask)
            dd = dfunc(a, b)
            assert isinstance(dd, np.ma.masked_array)
            self.assertEqual(list(dd), [0.5, 1.0, np.ma.masked])


if __name__ == "__main__":
    unittest.main()
